'''Handles recovery and restarting long running tasks.'''

"""
    This module provides a class that manages the checkpointing of long running tasks.
    The CheckpointManager is constructued by the client providing a path to a CSV file
    (which may be non-existent).
    
    For each checkpoint, the client provides a DataFrame with the specified column. CheckpointManager
    updates the CSV file with the new data.

    For recovery, CheckpointManager reads the CSV file and returns a dataframe of the CSV.
 """

import numpy as np
import os  # type: ignore
import pandas as pd  # type: ignore
from typing import List


class CheckpointManager(object):
 
    def __init__(self, path:str, is_report:bool=True, is_initialize:bool=False)->None:
        """
        Args:
            path (str): Path to the CSV file
            is_report (bool): If True, reports progress
            is_initialize(bool): If True, deletes any existing checkpoint file
        """
        self.path = path
        self.is_report = is_report
        #
        self._print(f"CheckpointManager: {self.path}")
        if os.path.exists(self.path):
            if is_initialize:
                os.remove(self.path)
                self._print(f"Deleted: {self.path}")
            else:
                self._print(f"Recovering from: {self.path}")
        else:
            self._print(f"Creating: {self.path}")

    def _print(self, msg:str)->None:
        if self.is_report:
            print(f"***{msg}")

    def checkpoint(self, df:pd.DataFrame)->None:
        """
        Checkpoints a DataFrame.
        """
        self._print(f"Checkpointing a dataframe of length {len(df)} to {self.path}")
        df.to_csv(self.path, index=False)

    def recover(self)->pd.DataFrame:
        """
        Recovers a previously saved DataFrame.

        Returns:
            np.ndarray: List of processed tasks
        """
        if not os.path.exists(self.path):
            df = pd.DataFrame()
        else:
            try:
                df = pd.read_csv(self.path)
            except:
                df = pd.DataFrame()
        self._print(f"Recovering a dataframe of length {len(df)} from {self.path}")
        return df
    
    def mergeCheckpoint(self, checkpoint_managers:List['CheckpointManager'])->int:
        """
        Merges the checkpoints from the checkpoint managers.

        Args:
            List (CheckpointManager): List of CheckpointManager objects

        Returns:
            int: Number of rows in the merged dataframe
        """
        dfs = [cm.recover() for cm in checkpoint_managers]
        df = pd.concat(dfs, ignore_index=False)
        self.checkpoint(df)
        return len(df)
    
    def remove(self):
        """Remove the checkpoint file."""
        if os.path.exists(self.path):
            os.remove(self.path)
            self._print(f"Deleted: {self.path}")


from pySubnetSB.significance_calculator_core import SignificanceCalculatorCore  # type: ignore
from pySubnetSB.network import Network  # type: ignore
import pySubnetSB.constants as cn # type: ignore

import numpy as np
import pandas as pd # type: ignore
import matplotlib.pyplot as plt
import unittest


IGNORE_TEST = False
IS_PLOT = False
SIMPLE_MODEL = """
S1 -> S2; k1*S1
S2 -> S3; k2*S2

S1 = 1
S2 = 0
S3 = 0
k1 = 0.1
k2 = 0.2
"""
COMPLEX_MODEL = """
S1 -> S2; k1*S1
S2 -> S3; k2*S2
S3 -> S1; k3*S3
S3 -> S2; k4*S3
S3 -> S4; k5*S3

S1 = 1
S2 = 0
S3 = 0
k1 = 0.1
k2 = 0.2
k3 = 0.2
k4 = 0.2
k5 = 0.2
"""
NUM_TARGET_NETWORK = 10
MAX_NUM_ASSIGNMENT = int(1e6)
NUM_TARGET_REACTION = 10
NUM_TARGET_SPECIES = 10
IDENTITY = cn.ID_STRONG
REFERENCE_NETWORK = Network.makeFromAntimonyStr(SIMPLE_MODEL)


#############################
# Tests
#############################
class TestSignificanceCalculator(unittest.TestCase):

    def setUp(self):
        self.calculator = SignificanceCalculatorCore(NUM_TARGET_SPECIES,
              NUM_TARGET_REACTION, num_target_network=NUM_TARGET_NETWORK)
        
    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertEqual(self.calculator.num_target_reaction, NUM_TARGET_REACTION)
        self.assertLessEqual(self.calculator.num_target_species, NUM_TARGET_SPECIES)
        self.assertGreaterEqual(self.calculator.num_target_network, 
            len(self.calculator.target_networks))
    
    def testCalculateSimple(self):
        if IGNORE_TEST:
            return
        for identity in cn.ID_LST:
            result = self.calculator.calculateSubnet(REFERENCE_NETWORK, max_num_assignment=MAX_NUM_ASSIGNMENT,
                is_report=IGNORE_TEST, identity=identity)
            self.assertTrue(result.num_reference_species > 0)
            self.assertTrue(result.num_reference_reaction > 0)
            self.assertEqual(result.num_target_species, NUM_TARGET_SPECIES)
            self.assertEqual(result.num_target_reaction, NUM_TARGET_REACTION)
            self.assertEqual(result.max_num_assignment, MAX_NUM_ASSIGNMENT)
            self.assertEqual(result.identity, identity)
            self.assertTrue(result.num_induced >= 0)
            self.assertTrue(result.num_truncated >= 0)
            self.assertTrue(result.frac_induced >= 0)
            self.assertTrue(result.frac_truncated >= 0)
    
    def testCalculateComplex(self):
        if IGNORE_TEST:
            return
        reference_network = Network.makeFromAntimonyStr(COMPLEX_MODEL)
        calculator = SignificanceCalculatorCore(NUM_TARGET_SPECIES,
              NUM_TARGET_REACTION, num_target_network=1000*NUM_TARGET_NETWORK)
        result = calculator.calculateSubnet(reference_network, max_num_assignment=MAX_NUM_ASSIGNMENT,
              is_report=IS_PLOT)
        self.assertTrue(result.frac_induced < 0.1)
    
    def testCalculateComplexEquality(self):
        if IGNORE_TEST:
            return
        for identity in cn.ID_LST:
            results = []
            for size in [2, 4]:
                reference_network = Network.makeRandomNetworkByReactionType(num_reaction=size,
                      num_species=size, is_exact=True)
                calculator = SignificanceCalculatorCore(size, size, num_target_network=1000)
                result = calculator.calculateEqual(reference_network,
                      max_num_assignment=MAX_NUM_ASSIGNMENT,
                      is_report=IS_PLOT, identity=identity)
                results.append(result)
            self.assertTrue(results[0].frac_induced >= results[1].frac_induced)
        

if __name__ == '__main__':
    unittest.main()


'''A ConstraintOption is a set of boolean options for constraints.'''

import pySubnetSB.constants as cn # type: ignore
import pySubnetSB.util as util  # type: ignore


import string  # type: ignore
from typing import List


#####################################
class ConstraintOptionCollection(object): 

    # Subclasses for species and reactions

    def __init__(self, **kwargs):
        """
        kwargs:
            names of boolean options and their initial values
        """
        self.__dict__.update(kwargs)
        self.option_names = list(kwargs.keys())
        self.long_to_short_dct = {n: self.short_names[i] for i, n in enumerate(self.option_names)}
        self.short_to_long_dct = {v: k for k, v in self.long_to_short_dct.items()}
        self.num_option = len(self.option_names)

    def copy(self)->'ConstraintOptionCollection':
        """
        Returns a copy of the current object.
        """
        dct = {k: self.__dict__[k] for k in self.option_names}
        new_option = self.__class__(**dct)
        return new_option
    
    def __eq__(self, other)->bool:
        # Returns True if all constraint options are the same.
        for option_name in self.option_names:
            if self.__dict__[option_name] != other.__dict__[option_name]:
                return False
        return True

    def __repr__(self)->str:
        return str(self.__class__) + ": " + ', '.join(self.getTrueNames())
    
    def getTrueNames(self)->list[str]:
        results:list = []
        for i, key in enumerate(self.option_names):
            if self.__dict__[key]:
                results.append(key)
        return results

    @property 
    def collection_short_name(self)->str:
        # Creates a short name for the collection based on the options that are True.
        names = self.getTrueNames()
        if len(names) == 0:
            return cn.NONE
        return '+'.join([self.long_to_short_dct[n] for n in names])
    
    @property 
    def long_name(self)->str:
        # Returns a combination of short names to describe the current state
        names = self.getTrueNames()
        if len(names) == 0:
            return cn.NONE
        return '+'.join(names)
    
    def setAllFalse(self)->None:
        """
        Set all values to False.
        """
        for option_name in self.option_names:
            self.__dict__[option_name] = False

    def makeOptionFromShortNames(self, short_names:List[str])->'ConstraintOptionCollection':
        """
        Returns a new ConstraintOption object with the short names set to True.

        Args:
            short_names: List[str] Short names of the options to set to True.
        """
        long_names = [self.short_to_long_dct[n] for n in short_names]
        new_option = self.copy()
        new_option.setAllFalse()
        for name in long_names:
            new_option.__dict__[name] = True
        return new_option
    
    def makeFromCollectionShortName(self, collection_short_name:str)->'ConstraintOptionCollection':
        """
        Returns a new ConstraintOptionCollection with the options specified by the collection short name.

        Args:
            collection_short_name (str): _description_

        Returns:
            ConstraintOptionCollection: _description_
        """
        option_collection = self.copy()
        short_names = collection_short_name.split('+')
        option_collection.setAllFalse()
        for short_name in short_names:
            option_collection.__dict__[self.short_to_long_dct[short_name]] = True
        return option_collection
    
    def isTrue(self, option_name:str)->bool:
        """
        Returns True if the option is True.

        Args:
            option_name (str): _description_

        Returns:
            bool: _description_
        """
        if not option_name in self.__dict__.keys():
            return False
        return self.__dict__[option_name]

    def iterator(self):
        """
        Iteratively returns an OptionCollection of the same type with all combinations of
        values for the boolean attributes.
        """
        name_collections = util.getAllSubsets(self.short_names)
        for name_collection in name_collections:
            yield self.makeOptionFromShortNames(name_collection)


#####################################
class ReactionConstraintOptionCollection(ConstraintOptionCollection): 

    def __init__(self,
                 #is_make_successor_predecessor_constraint_matrix:bool=True,
                 is_make_n_step_constraint_matrix:bool=True,
                 is_make_classification_constraint_matrix:bool=True,
                 #is_make_autocatalysis_constraint_matrix:bool=True,
                 ):
        self.short_names = list(string.ascii_letters[:26])
        self.short_names.reverse()
        self.short_names = self.short_names[:2]
        self.short_names.reverse()
        super().__init__(
            #is_make_successor_predecessor_constraint_matrix=is_make_successor_predecessor_constraint_matrix,
            is_make_n_step_constraint_matrix=is_make_n_step_constraint_matrix,
            is_make_classification_constraint_matrix=is_make_classification_constraint_matrix,
            #is_make_autocatalysis_constraint_matrix=is_make_autocatalysis_constraint_matrix,
            )
        

#####################################
class SpeciesConstraintOptionCollection(ConstraintOptionCollection): 


    def __init__(self, 
                 #is_reactant_product_count_constraint_matrix:bool=True,
                 #is_autocatalysis_constraint:bool=True,
                 is_reactant_product_constraint_matrix:bool=True,
                 #is_successor_predecessor_constraint_matrix:bool=True,
                 is_n_step_constraint_matrix:bool=True):
        self.short_names = list(string.ascii_letters[:2])
        super().__init__(
              #is_reactant_product_count_constraint_matrix=is_reactant_product_count_constraint_matrix,
              #is_autocatalysis_constraint=is_autocatalysis_constraint,
              #is_successor_predecessor_constraint_matrix=is_successor_predecessor_constraint_matrix,
              is_n_step_constraint_matrix=is_n_step_constraint_matrix,
              is_reactant_product_constraint_matrix=is_reactant_product_constraint_matrix,
        )


from pySubnetSB.significance_calculator import SignificanceCalculator  # type: ignore
from pySubnetSB.network import Network  # type: ignore
import pySubnetSB.constants as cn # type: ignore

import numpy as np
import pandas as pd # type: ignore
import matplotlib.pyplot as plt
import unittest


IGNORE_TEST = False
IS_PLOT = False
SIMPLE_MODEL = """
S1 -> S2; k1*S1
S2 -> S3; k2*S2

S1 = 1
S2 = 0
S3 = 0
k1 = 0.1
k2 = 0.2
"""
COMPLEX_MODEL = """
S1 -> S2; k1*S1
S2 -> S3; k2*S2
S3 -> S1; k3*S3
S3 -> S2; k4*S3
S3 -> S4; k5*S3

S1 = 1
S2 = 0
S3 = 0
k1 = 0.1
k2 = 0.2
k3 = 0.2
k4 = 0.2
k5 = 0.2
"""
NUM_ITERATION = 10
MAX_NUM_ASSIGNMENT = int(1e6)
NUM_TARGET_REACTION = 10
NUM_TARGET_SPECIES = 10
IDENTITY = cn.ID_STRONG
REFERENCE_NETWORK = Network.makeFromAntimonyStr(SIMPLE_MODEL)


#############################
# Tests
#############################
class TestSignificanceCalculator(unittest.TestCase):

    def setUp(self):
        self.calculator = SignificanceCalculator(REFERENCE_NETWORK, NUM_TARGET_SPECIES,
              NUM_TARGET_REACTION, identity=IDENTITY)
        
    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertTrue(self.calculator.reference_network is not None)
        self.assertLessEqual(self.calculator.num_target_reaction, NUM_TARGET_REACTION)
        self.assertLessEqual(self.calculator.num_target_species, NUM_TARGET_SPECIES)
        self.assertEqual(self.calculator.identity, IDENTITY)
    
    def testCalculateSimple(self):
        if IGNORE_TEST:
            return
        result = self.calculator.calculate(NUM_ITERATION, max_num_assignment=MAX_NUM_ASSIGNMENT,
              is_report=IGNORE_TEST)
        self.assertTrue(result.num_reference_species > 0)
        self.assertTrue(result.num_reference_reaction > 0)
        self.assertEqual(result.num_target_species, NUM_TARGET_SPECIES)
        self.assertEqual(result.num_target_reaction, NUM_TARGET_REACTION)
        self.assertEqual(result.max_num_assignment, MAX_NUM_ASSIGNMENT)
        self.assertEqual(result.identity, IDENTITY)
        self.assertTrue(result.num_induced >= 0)
        self.assertTrue(result.num_truncated >= 0)
        self.assertTrue(result.frac_induced >= 0)
        self.assertTrue(result.frac_truncated >= 0)
    
    def testCalculateComplex(self):
        if IGNORE_TEST:
            return
        reference_network = Network.makeFromAntimonyStr(COMPLEX_MODEL)
        calculator = SignificanceCalculator(reference_network, NUM_TARGET_SPECIES,
              NUM_TARGET_REACTION, identity=IDENTITY)
        result = calculator.calculate(NUM_ITERATION, max_num_assignment=MAX_NUM_ASSIGNMENT,
              is_report=False)
        self.assertTrue(result.frac_induced < 0.1)

    def testPlotSignificance(self):
        # Plots probability of induced network in a random target as the number of iterations increases
        if IGNORE_TEST:
            return
        result = self.calculator.plotSignificance(
              is_report=IGNORE_TEST,
              num_iteration=5, is_plot=False)
        for values in [result.frac_induces, result.frac_truncates]:
            self.assertTrue(len(result.target_sizes), len(values))

    def testCalculateOccurrenceProbability(self):
        if IGNORE_TEST:
            return
        reference_network = REFERENCE_NETWORK
        reference_network = Network.makeFromAntimonyStr(COMPLEX_MODEL)
        result = self.calculator.calculateNetworkOccurrenceProbability(
              reference_network, num_iteration=NUM_ITERATION, is_report=False,
              max_num_assignment=MAX_NUM_ASSIGNMENT)
        self.assertTrue(result[0] < 0.01)

    def testPlotProbabilityOfOccurrence(self):
        if IGNORE_TEST:
            return
        size = 100
        num_reactions = np.random.randint(1, 10, size=size)
        num_species = np.random.randint(1, 10, size=size)
        values = np.random.uniform(0, 1, size=size)
        df = pd.DataFrame({'num_reaction': num_reactions, 'num_species': num_species,
                           'value': values})
        self.calculator.plotProbabilityOfOccurrence(df, column= 'value', is_plot=IS_PLOT)
        

if __name__ == '__main__':
    unittest.main()


'''Calculates the reactant, product, net stoichiometry matrices.'''
# Originally developed by Herbert M Sauro, University of Washington.

import tellurium as te  # type: ignore
import simplesbml  # type: ignore
import numpy as np  # type: ignore
from typing import Any, Optional

# Result: namedtuple
#  reactant_mat: reactant stoichiometry matrix
#  product_mat: product stoichiometry matrix
#  standard_mat: the usual stoichiometry matrix
#  species_names: list of species names
#  reaction_names: list of reaction names


class Stoichiometry(object):

    def __init__(self, antimony_str:str, roadrunner:Optional[Any]=None)->None:

        """
        Calculate the stoichiometry matrix from an SBML model.
        Args:
            antimony_str (str): Antimony model
        Returns:
            Result: Result(reactant_mat, product_mat, stoichiometry_mat, species_names, reaction_names)
        """
        self.antimony_str = antimony_str
        self.roadrunner = roadrunner
        self.reactant_mat, self.product_mat, self.standard_mat, self.species_names, self.reaction_names \
            = self.calculate() 
        
    def calculate(self):
        if self.roadrunner is None:
            self.roadrunner = te.loada(self.antimony_str)
        #roadrunner.conservedMoietyAnalysis = True
        sbml = self.roadrunner.getSBML()
        try:
            model = simplesbml.loadSBMLStr(sbml)
        except:
            return None, None, None, None, None
        # Model inforeactant_mation
        num_floating_species = model.getNumFloatingSpecies()
        num_boundary_species = model.getNumBoundarySpecies()
        num_species = num_floating_species + num_boundary_species
        num_reaction = model.getNumReactions()
        species_names = [model.getNthFloatingSpeciesId(i) for i in range(num_floating_species)]
        species_names.extend([model.getNthBoundarySpeciesId(i) for i in range(num_boundary_species)])
        reaction_names = [model.getNthReactionId(i) for i in range(num_reaction)]

        # Allocate space for the stoichiometry matrix
        reactant_mat = np.zeros((num_species, num_reaction))
        product_mat = np.zeros((num_species, num_reaction))
        standard_mat = np.zeros((num_species, num_reaction))
        for ispecies in range(num_species):
            if ispecies < num_floating_species:
                speciesId = model.getNthFloatingSpeciesId(ispecies)
            else:
                speciesId = model.getNthBoundarySpeciesId(ispecies - num_floating_species)
            for ireaction in range(num_reaction):
                # Get the product stoichiometry
                numProducts = model.getNumProducts(ireaction)
                for k1 in range(numProducts):
                    productId = model.getProduct(ireaction, k1)
                    if speciesId == productId:
                        product_mat[ispecies, ireaction] += model.getProductStoichiometry(ireaction, k1)
                # Get the reactant stoihiometry
                numReactants = model.getNumReactants(ireaction)
                for k1 in range(numReactants):
                    reactantId = model.getReactant(ireaction, k1)
                    #reactionId = model.getNthReactionId(ireaction)
                    if speciesId == reactantId:
                        reactant_mat[ispecies, ireaction] += model.getReactantStoichiometry(ireaction, k1)
        # Calculate the stoichiometry matrix
        standard_mat = product_mat - reactant_mat
        num_row, num_column = standard_mat.shape
        if (num_row != len(species_names)) or (num_column != len(reaction_names)):
            raise RuntimeError("The stoichiometry matrix is not the correct size!")
        # return result
        return reactant_mat, product_mat, standard_mat, species_names, reaction_names



'''Calculates the significance level of finding an induced subnetwork in a target network.'''

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.network import Network  # type: ignore

import collections
import numpy as np
import tqdm # type: ignore
from typing import List, Tuple, Optional


NUM_REFERENCE_SPECIES = "num_reference_species"
NUM_REFERENCE_REACTION = "num_reference_reaction"
NUM_TARGET_SPECIES = "num_target_species"
NUM_TARGET_REACTION = "num_target_reaction"
FRAC_INDUCED = "frac_induced"
FRAC_TRUNCATED = "frac_truncated"
# Default values
NUM_ITERATION = 1000


SignificanceCalculatorCoreResult = collections.namedtuple("SignificanceCalculatorCoreResult", 
  ["num_reference_species", "num_reference_reaction", "num_target_species", "num_target_reaction",
   "num_target_network", "max_num_assignment", "identity",
   "num_induced", "num_truncated", "frac_induced", "frac_truncated"])


class SignificanceCalculatorCore(object):

    def __init__(self, num_target_species:int, num_target_reaction:int, num_target_network:int):
        self.num_target_reaction = num_target_reaction
        self.num_target_species = num_target_species
        self.num_target_network = num_target_network
        target_networks = [Network.makeRandomNetworkByReactionType(num_target_reaction,
                num_target_species, is_exact=False) for _ in range(num_target_network)]
        self.target_networks = [n for n in target_networks
                if (n.num_species == num_target_species) and (n.num_reaction == num_target_reaction)]
        self.count_target_network = len(self.target_networks)
        self.target_dct:Optional[dict] = None

    def calculateSubnet(self, reference_network:Network, identity:str=cn.ID_WEAK, is_subnet:bool=True,
          max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT, is_report:bool=True
          )->SignificanceCalculatorCoreResult:
        """
        Calculates the significance level of finding an induced subnetwork in a target network.

        Args:
            reference_network (Network): Reference network
            is_subnet (bool): If True, the target network is a subnet of the reference network
            identity (str): Identity type
            max_num_assignment (int): Maximum number of assignment pairs
            is_report (bool): If True, report progress

        Returns:
            SignificanceCalculatorCoreResult
        """
        num_induced = 0
        num_truncated = 0
        for target_network in tqdm.tqdm(self.target_networks, desc="iteration", disable=not is_report):
            result = reference_network.isStructurallyIdentical(target_network,
                  identity=identity, is_subnet=is_subnet,
                    max_num_assignment=max_num_assignment, is_report=False)
            num_induced += 1 if result else 0
            num_truncated += 1 if result.is_truncated else 0
        # Calculate the significance level
        return SignificanceCalculatorCoreResult(
            num_reference_species=reference_network.num_species,
            num_reference_reaction=reference_network.num_reaction,
            num_target_species=self.num_target_species,
            num_target_reaction=self.num_target_reaction,
            num_target_network=self.count_target_network,
            max_num_assignment=max_num_assignment,
            identity=identity,
            num_induced=num_induced,
            num_truncated=num_truncated,
            frac_induced=num_induced/self.count_target_network if self.count_target_network > 0 else np.nan,
            frac_truncated=num_truncated/self.count_target_network if self.count_target_network > 0 else np.nan)

    def calculateEqual(self, reference_network:Network, identity:str=cn.ID_WEAK,
          max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT, is_report:bool=True
          )->SignificanceCalculatorCoreResult:
        """
        Calculates the significance level of finding a structurally identical target network.

        Args:
            reference_network (Network): Reference network
            identity (str): Identity type
            max_num_assignment (int): Maximum number of assignment pairs
            is_report (bool): If True, report progress

        Returns:
            SignificanceCalculatorCoreResult
        """
        if self.target_dct is None:
            self.target_dct = {}
            for target_network in self.target_networks:
                key = target_network.network_hash
                if not key in self.target_dct:
                    self.target_dct[key] = []
                self.target_dct[key].append(target_network)
        # Look for structurally identical networks
        num_induced = 0
        num_truncated = 0
        key = reference_network.network_hash
        if key in self.target_dct:
            for target_network in tqdm.tqdm(self.target_dct[key], desc="iteration", disable=not is_report):
                result = reference_network.isStructurallyIdentical(target_network, identity=identity,
                    is_subnet=False,
                    max_num_assignment=max_num_assignment, is_report=False)
                num_induced += 1 if result else 0
                num_truncated += 1 if result.is_truncated else 0
        # Return the result
        if self.count_target_network > 0:
            frac_induced = num_induced/self.count_target_network
            frac_truncated = num_truncated/self.count_target_network
        else:
            frac_induced = np.nan
            frac_truncated = np.nan
        return SignificanceCalculatorCoreResult(
            num_reference_species=reference_network.num_species,
            num_reference_reaction=reference_network.num_reaction,
            num_target_species=self.num_target_species,
            num_target_reaction=self.num_target_reaction,
            num_target_network=self.count_target_network,
            max_num_assignment=max_num_assignment,
            identity=identity,
            num_induced=num_induced,
            num_truncated=num_truncated,
            frac_induced=frac_induced,
            frac_truncated=frac_truncated,
        )


'''Creates workunits for workers, starts Worker processes, combines Worker results in their checkpoint files.'''

"""
Parallelism is achived by running workers in parallel. 
"""


import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.model_serializer import ModelSerializer # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.parallel_subnet_finder_worker import executeWorker, Workunit # type: ignore
from pySubnetSB.checkpoint_manager import CheckpointManager # type: ignore
from pySubnetSB.worker_checkpoint_manager import WorkerCheckpointManager # type: ignore
from pySubnetSB.network import Network # type: ignore
from pySubnetSB.subnet_finder_workunit_manager import SubnetFinderWorkunitManager  # type: ignore
from pySubnetSB import util

import itertools
import os
import multiprocessing as mp
import time
from concurrent.futures import ProcessPoolExecutor
import numpy as np
import pandas as pd  # type: ignore
from typing import List, Optional

CHECKPOINT_FILE = "parallel_subset_finder_checkpoint.txt"
WORKUNIT_CSV_FILE = "parallel_subset_finder_workunit.csv"
SERIALIZATION_FILE = "serialized.txt"
REFERENCE_SERIALIZATION_FILENAME = "reference_serialized.txt"
TARGET_SERIALIZATION_FILENAME = "target_serialized.txt"
CHECKPOINT_PATH = os.path.join(cn.AUXILIARY_DATA_DIR, CHECKPOINT_FILE)
WORKUNIT_CSV_PATH = os.path.join(cn.AUXILIARY_DATA_DIR, WORKUNIT_CSV_FILE)
CHECKPOINT_INTERVAL = 100 # Number of workunits procesed between checkpoints
BIOMODELS_SERIALIZATION_TARGET_FILENAME = 'biomodels_serialized_target.txt'  # Serialized target models
BIOMODELS_SERIALIZATION_REFERENCE_FILENAME = 'biomodels_serialized_reference.txt' # Serialized reference models
BIOMODELS_CHECKPOINT_FILENAME = "biomodels_checkpoint.csv"
BIOMODELS_CHECKPOINT_PATH = os.path.join(cn.AUXILIARY_DATA_DIR, BIOMODELS_CHECKPOINT_FILENAME)

# Performance monitoring
LAST_TIME = "last_time"
global_dct = {LAST_TIME: time.time()}
#####
def printPerformanceInformation(msg:str, initialize:bool=False):
    # Print performance information
    #if True and not initialize:
    if False:
        last_time = global_dct["last_time"]
        now_time = time.time()
        duration = now_time - last_time
        new_msg = f"**{msg}: {duration} seconds."
        print(new_msg, flush=True)
    global_dct[LAST_TIME]= time.time()
####


############################### CLASSES ###############################
class ParallelSubnetFinder(object):
    # Finds subnets of target models for reference models

    def __init__(self, reference_serialization_path:str, target_serialization_path:str,
          identity:str=cn.ID_WEAK,
          checkpoint_path:str=CHECKPOINT_PATH,
          workunit_csv_path:str=WORKUNIT_CSV_PATH)->None:
        """
        Args:
            reference_serialization_path (str): Path to the reference model serialization file
            target_serialization_path (str): Path to the target model serialization file
            identity (str): Identity type
            checkpoint_path (str): Path to the output serialization file that checkpoints results
               for all tasks.
            workunit_csv_path (str): Path to the CSV file that contains workunits. Overwritten.
        """
        self.reference_serialization_path = reference_serialization_path
        serializer = ModelSerializer(None, reference_serialization_path)
        self.reference_networks = serializer.deserialize().networks
        self.target_serialization_path = target_serialization_path
        serializer = ModelSerializer(None, target_serialization_path)
        self.target_networks = serializer.deserialize().networks
        self.identity = identity
        self.checkpoint_path = checkpoint_path
        self.workunit_csv_path = workunit_csv_path
    
    @classmethod
    def findFromNetworks(cls, reference_networks:List[Network],
          target_networks:List[Network],
          identity:str=cn.ID_WEAK,
          is_report:bool=True,
          serialization_dir:str=cn.DATA_DIR,
          is_initialize:bool=False,
          num_worker:int=-1,
          max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT,
          reference_serialization_filename:str=REFERENCE_SERIALIZATION_FILENAME,
          target_serialization_filename:str=TARGET_SERIALIZATION_FILENAME,
          checkpoint_path:str=CHECKPOINT_PATH)->pd.DataFrame:
        """
        Finds subnets of SBML/Antmony models in a target directory for SBML/Antimony models in a reference directory.

        Args:
            reference_networks (List[Network]): List of reference networks
            target_networks (List[Network]): List of target networks
            identity (str): Identity type
            serialization_dir (str): Directory for serialization files for reference and target
            is_report (bool): If True, report progress
            is_initialize (bool): If True, initialize the checkpoint
            num_worker (int): Number of workers/CPUs to run (<0 is all CPUs)
            max_num_assignment (int): Maximum number of assignment pairs
            reference_serialization_filename (str): Filename for the reference serialization file
            target_serialization_filename (str): Filename for the target serialization file
            checkpoint_path (str): Path to the output serialization file that checkpoints results

        Returns:
            pd.DataFrame: (See parallelFind)
        """
        #####
        def serializeNetworks(networks, filename:str)->str:
            """
            Serializes the networks to a file, returning the serialization path.

            Args:
                networks (List[Network]): List of networks
                filename (str): Filename

            Returns:
                str: Serialization path
            """
            serialization_path = os.path.join(serialization_dir, filename)
            serializer = ModelSerializer(serialization_dir, serialization_path)
            serializer.serializeNetworks(networks, is_initialize=True)
            return serialization_path
        #####
        target_serialization_path = serializeNetworks(target_networks, target_serialization_filename)
        reference_serialization_path = serializeNetworks(reference_networks,
              reference_serialization_filename)
        #
        finder = cls(reference_serialization_path, target_serialization_path,
          identity=identity, checkpoint_path=checkpoint_path)
        return finder.parallelFind(is_report=is_report, is_initialize=is_initialize,
              max_num_assignment=max_num_assignment, num_worker=num_worker)
    
    @classmethod
    def directoryFind(cls, reference_directory, target_directory, identity:str=cn.ID_WEAK,
          checkpoint_path:str=CHECKPOINT_PATH, num_worker:int=-1, is_initialize:bool=False,
          max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT,
          is_report:bool=True)->pd.DataFrame:
        """ 
        Finds subnets of SBML/Antmony models in a target directory for SBML/Antimony models in a reference directory.

        Args:
            reference_directory (str): Directory that contains the reference model files (or serialization path)
            target_directory (str): Directory that contains the target model files or a serialization file (".txt")
            identity (str): Identity type
            num_worker (int): Number of workers
            checkpoint_path (str): Path to the output serialization file that checkpoints results
            is_initialize (bool): If True, initialize the checkpoint
            is_report (bool): If True, report progress

        Returns:
            pd.DataFrame: (See find)
        """
        #####
        def getSerializationPath(directory:str)->str:
            """ 
            Obtains the networks from a directory or serialization file.

            Args:
                directory (str): directory path or path to serialization file

            Returns:
                serialization path
            """
            if directory.endswith(".txt"):
                serialization_path = directory
            else:
                # Construct the serialization file path and file
                serialization_path = os.path.join(directory, SERIALIZATION_FILE)
                if not os.path.exists(serialization_path):
                    serializer = ModelSerializer(directory, serialization_path)
                    serializer.serialize()
            return serialization_path
        #####
        reference_serialization_path = getSerializationPath(reference_directory)
        target_serialization_path = getSerializationPath(target_directory)
        # Put the serialized models in the directory. Check for it on invocation.
        finder = cls(reference_serialization_path, target_serialization_path,
          identity=identity, checkpoint_path=checkpoint_path)
        return finder.parallelFind(is_report=is_report, is_initialize=is_initialize,
              num_worker=num_worker, max_num_assignment=max_num_assignment)
    
    def parallelFind(self, num_worker:int=-1, is_report:bool=True, is_initialize:bool=True,
          max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT)->pd.DataFrame:
        """
        Finds reference networks that are subnets of the target networks using parallel processing.

        Args:
            is_report (bool): If True, report progress
            is_initialize (bool): If True, initialize the checkpoint
            max_num_assignment (int): Maximum number of assignments to process

        Returns:
            pd.DataFrame: (See cn.FINDER_DATAFRAME_COLUMNS)
        """
        #####
        def _print(msg:str):
            if is_report:
                print(msg)
        #####        
        def getArgs(worker_idx:int, is_terminate:bool=True)->tuple:
            is_initialize_worker = True  # Initialize the worker checkpoint
            return worker_idx, self.workunit_csv_path, num_worker, self.checkpoint_path, \
                          self.reference_serialization_path, self.target_serialization_path, \
                          self.identity, is_report, \
                          is_initialize_worker, \
                          max_num_assignment,  \
                          CHECKPOINT_INTERVAL, is_terminate
        #####
        printPerformanceInformation("parallelFind/time0", initialize=True)
        # Initializations
        if num_worker < 0:
            num_worker = mp.cpu_count()
        # Recover checkpoint results
        manager = CheckpointManager(self.checkpoint_path, is_initialize=is_initialize,
              is_report=is_report)
        df = manager.recover()
        if len(df) > 0:
            known_names = set(df[cn.FINDER_REFERENCE_NAME]).intersection([n.network_name for n in self.reference_networks])
            df = df[df[cn.FINDER_REFERENCE_NAME].isin(known_names)]
        reference_dct = {n.network_name: idx for idx, n in enumerate(self.reference_networks)}
        target_dct = {n.network_name: idx for idx, n in enumerate(self.target_networks)}
        if len(df) > 0:
            ref_idx_arr = np.array([reference_dct[str(n)] for n in df[cn.FINDER_REFERENCE_NAME]])
            tgt_idx_arr = np.array([target_dct[str(n)] for n in df[cn.FINDER_TARGET_NAME]])
            completed_encodes = set(util.encodeIntPair(ref_idx_arr, tgt_idx_arr))
        else:
            completed_encodes = set([])
        all_pairs = list(itertools.product(range(len(self.reference_networks)),
              range(len(self.target_networks))))
        all_ref_idx_arr = np.array([i for i, j in all_pairs])
        all_tgt_idx_arr = np.array([j for i, j in all_pairs])
        all_pair_encodeds = set(util.encodeIntPair(all_ref_idx_arr, all_tgt_idx_arr))
        unfinished_arr = np.array(list(all_pair_encodeds - completed_encodes))
        unfinished_pairs = util.decodeIntPair(unfinished_arr)
        workunits = [Workunit(i, j) for i, j in zip(unfinished_pairs[0], unfinished_pairs[1])]
        printPerformanceInformation(f"parallelFind/time1: {len(workunits)} workunits.")
        if len(workunits) > 0:
            # Create the workunits
            workunit_manager = SubnetFinderWorkunitManager(self.workunit_csv_path, num_worker=num_worker,
                   reference_networks=self.reference_networks, target_networks=self.target_networks)
            workunit_manager.makeWorkunitFile()
            if num_worker == 1:
               # Handle single process
               args = getArgs(0, is_terminate=False)
               executeWorker(*args)
            else:
               # Parallel processing
               processes = []
               for process_idx in range(num_worker):
                   proc = mp.Process(target=executeWorker, args=(getArgs(process_idx)))
                   processes.append(proc)
                   proc.start()
               # Wait for the processes to finish
               for process_idx, proc in enumerate(processes):
                   proc.join()
                   printPerformanceInformation(f"parallelFind/time3a, process {process_idx}")
            # Wait for workers to complete
            _print(f"**{num_worker} processes completed.")
            printPerformanceInformation("parallelFind/time4")
            # Merge the results
            merged_checkpoint_result = WorkerCheckpointManager.merge(
                  self.checkpoint_path, num_worker, is_report=is_report)
            printPerformanceInformation(f"parallelFind/time5")
            df = merged_checkpoint_result.merged_checkpoint_manager.recover().full_df
            # Eliminate the checkpoint files
            for worker_idx in range(num_worker):
                worker_checkpoint_path = WorkerCheckpointManager.makeWorkerCheckpointPath(
                    self.checkpoint_path, worker_idx)
                worker_checkpoint_manager = WorkerCheckpointManager(worker_checkpoint_path)
                worker_checkpoint_manager.remove()
        else:
            # No workunits to process
            df = manager.recover()
        # Return the results
        if is_report:
            msg = f"**Done. Processed {len(df)}"
            msg += f" reference networks in {self.checkpoint_path}."
            print(msg)
        return df
    
    @classmethod
    def biomodelsFind(cls,
          reference_network_size:int=15,
          max_num_reference_network:int=-1,
          max_num_target_network:int=-1,
          identity:str=cn.ID_STRONG,
          max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT,
          reference_network_names:List[str]=[],
          is_report:bool=True,
          serialization_dir:str=cn.DATA_DIR,
          reference_serialization_filename:str=REFERENCE_SERIALIZATION_FILENAME,
          target_serialization_filename:str=TARGET_SERIALIZATION_FILENAME,
          checkpoint_path:Optional[str]=None,
          num_worker:int=-1,
          is_no_boundary_network:bool=True,
          skip_networks:Optional[List[str]]=None,
          is_initialize:bool=True)->pd.DataFrame:
        """
        Finds subnets of BioModels models for reference models specified by a maximum size.
        The DataFrame returned includes a reference network, target network pair with null strings for found networks
        so that there is a record of all comparisons made.

        Args:
            reference_network_size (int): Size of networks in Bionetworks that are used as reference network
            max_num_reference_network (int): Maximum number of reference networks to process. If -1, process all
            max_num_target_network (int): Maximum number of target networks to process. If -1, process all
            reference_network_names (List[str]): List of reference network names to process
            is_report (bool): If True, report progress
            serialization_dir (str): Directory for serialization files
            checkpoint_path (Optional[str]): Path for the checkpoint file
            num_worker (int): Number of workers to process the reference networks (-1 is all CPUs)
            skip_networks (List[str]): List of reference network to skip
            is_filter_unitnetworks (bool): If True, remove networks that only have reactions with one species
            is_initialize (bool): If True, initialize the checkpoint

        Returns:
            pd.DataFrame: (See find)
        """
        if checkpoint_path is None:
            checkpoint_path = BIOMODELS_CHECKPOINT_PATH
        # Select the networks to be analyzed
        if max_num_reference_network == -1:
            max_num_reference_network = int(1e9)
        if max_num_target_network == -1:
            max_num_target_network = int(1e9)
        # Construct the reference and target networks
        serializer = ModelSerializer(None, cn.BIOMODELS_SERIALIZATION_PATH)
        collection = serializer.deserialize()
        all_networks = collection.networks
        if skip_networks is not None:
            all_networks = [n for n in all_networks if n.network_name not in skip_networks]
        if is_no_boundary_network:
            all_networks = [n for n in all_networks if not Network.isBoundaryNetwork(n)]
        if len(reference_network_names) > 0:
            reference_networks = [n for n in all_networks if n.network_name in reference_network_names]
            if len(reference_networks) != len(reference_network_names):
                missing_networks = set(reference_network_names) - set([n.network_name for n in reference_networks])
                raise ValueError(f"Could not find reference networks {missing_networks}.")
        else:
            reference_networks = [n for n in all_networks if n.num_reaction <= reference_network_size]
            num_reference_network = min(len(reference_networks), max_num_reference_network)
            reference_networks = reference_networks[:num_reference_network]
        target_networks = [n for n in all_networks if n.num_reaction > reference_network_size]
        num_target_network = min(len(target_networks), max_num_target_network)
        target_networks = target_networks[:num_target_network]
        # Process the selected networks
        return cls.findFromNetworks(
          reference_networks,
          target_networks,
          identity=identity,
          is_report=is_report,
          serialization_dir=serialization_dir,
          reference_serialization_filename=reference_serialization_filename,
          target_serialization_filename=target_serialization_filename,
          is_initialize=is_initialize,
          num_worker=num_worker,
          max_num_assignment=max_num_assignment,
          checkpoint_path=checkpoint_path)


if __name__ == "__main__":
    df = ParallelSubnetFinder.biomodelsFind(reference_network_size=10, is_report=True,
          skip_networks=[], num_worker=-1, is_initialize=False)
    df.to_csv(os.path.join(cn.DATA_DIR, "biomodels_subnet.csv"))
    print("Done")


'''A container for structurally identical networks and their statistics. Has a string representation.'''

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB import constants as cn  # type: ignore
from pySubnetSB.assignment_pair import AssignmentPair  # type: ignore

import json
import time
from typing import Union, List, Optional


class ProcessedNetwork(object):

    def __init__(self, network_name:Union[Network, str])->None:
        self.network_name = self.convertToNetworkName(network_name)
        # Calculated
        self.is_indeterminate:bool = False
        self.processing_time:float = 0
        self.assignment_collection:List[AssignmentPair] = []

    def __repr__(self)->str:
        return f"{self.network_name}_{self.processing_time}_{self.is_indeterminate}_{self.assignment_collection}"

    @staticmethod 
    def convertToNetworkName(network:Union[Network, str])->str:
        """Returns the network name

        Args:
            network (Union[Network, str]): _description_

        Returns:
            str: _description_
        """
        if isinstance(network, str):
            network_name = network
        else:
            network_name = network.network_name
        return network_name

    def setIndeterminate(self, value)->None:
        self.is_indeterminate = value

    def setAssignmentCollection(self, assignment_collection:List[AssignmentPair])->None:
        self.assignment_collection = assignment_collection

    def addProcessingTime(self, processing_time:float)->None:
        """Adds to processing time

        Args:
            processing_time (Optional[float], optional): _description_. Defaults to None.
        """
        self.processing_time += processing_time

    def __eq__(self, other:object)->bool:
        if not isinstance(other, ProcessedNetwork):
            return False
        return (self.network_name == other.network_name and
                self.is_indeterminate == other.is_indeterminate)
    
    def copy(self)->'ProcessedNetwork':
        processed_network = ProcessedNetwork(self.network_name)
        processed_network.processing_time = self.processing_time
        processed_network.is_indeterminate = self.is_indeterminate
        processed_network.assignment_collection = [a.copy() for a in self.assignment_collection]
        return processed_network
    
#    def __repr__(self)->str:
#        repr_str = self._CSV_MAKER.encode(
#            network_name=self.network_name,
#            processing_time=self.processing_time,
#            is_indeterminate=self.is_indeterminate,
#            assignment_collection=self.assignment_collection)
#        return repr_str

#    @classmethod
#    def makeFromRepr(cls, repr_str:str)->'processedNetwork':
#        """
#        Constructs a processedNetwork from a string representation.
#
#        Args:
#            repr_str (str): _description_
#
#        Returns:
#            processedNetwork
#        """
#        dct = cls._CSV_MAKER.decode(repr_str)
#        processed_network = processedNetwork(dct[NETWORK_NAME])
#        processed_network.setAssignmentCollection(dct[ASSIGNMENT_COLLECTION])
#        processed_network.setProcessingTime(dct[PROCESSING_TIME])
#        processed_network.setIndeterminate(dct[IS_INDETERMINATE])
#        return processed_network
    
    def serialize(self)->str:
        """Creates a JSON string for the object.

        Returns:
            str
        """
        assignment_collection = [a.serialize() for a in self.assignment_collection]
        dct = {cn.S_ID: self.__class__.__name__,
               cn.S_NETWORK_NAME: self.network_name,
               cn.S_PROCESSING_TIME: self.processing_time,
               cn.S_IS_INDETERMINATE: self.is_indeterminate,
               cn.S_ASSIGNMENT_COLLECTION: assignment_collection}
        return json.dumps(dct)
    
    @classmethod
    def deserialize(cls, serialization_str)->'ProcessedNetwork':
        """Creates a processed network from a JSON serialization string.

        Args:
            serialization_str

        Returns:
            processedNetwork
        """
        dct = json.loads(serialization_str)
        if not cls.__name__ in dct[cn.S_ID]:
            raise ValueError(f"Expected {cls} but got {dct[cn.S_ID]}")
        processed_network = cls(dct[cn.S_NETWORK_NAME])
        processed_network.addProcessingTime(dct[cn.S_PROCESSING_TIME])
        processed_network.setIndeterminate(dct[cn.S_IS_INDETERMINATE])
        processed_network.setAssignmentCollection(dct[cn.S_ASSIGNMENT_COLLECTION])
        return processed_network


'''Container of networks. Can serialize and deserialize to a DataFrame. Construct from Antimony files. '''


from pySubnetSB import constants as cn   # type: ignore
from pySubnetSB.network import Network   # type: ignore

import json
import collections
import copy
import os
import pandas as pd # type: ignore
import numpy as np
import tellurium as te  # type: ignore
from typing import List, Optional, Dict


ANTIMONY_EXTS = [".ant", ".txt"]  # Antimony file extensions


ArrayContext = collections.namedtuple('ArrayContext', "string, num_row, num_column")


####################################
class NetworkCollection(object):
        
    def __init__(self, networks: List[Network], directory:Optional[str]=None)->None:
        """
        Args:
            networks (List[Network]): Networks in the collection
        """
        self.networks = networks
        self.network_dct = {n.network_name: n for n in networks}
        self.directory = directory

    def add(self, network:Network)->None:
        self.networks.append(network)

    def __eq__(self, other:'NetworkCollection')->bool:  # type: ignore
        # Check that collections have networks with the same attribute values
        if len(self) != len(other):
            return False
        # Check the network names
        network1_dct = {n.network_name: n for n in self.networks}
        network2_dct = {n.network_name: n for n in other.networks}
        key_set = set(network1_dct.keys())
        key_diff = key_set.symmetric_difference(set(network2_dct.keys()))
        if len(key_diff) > 0:
            return False
        #
        for key in key_set:
            if not network1_dct[key] == network2_dct[key]:
                return False
        return True

    def __len__(self)->int:
        return len(self.networks)
    
    def __repr__(self)->str:
        names = [n.network_name for n in self.networks]
        return "---".join(names)
    
    def _findCommonType(self, collection_identity_type1:str, collection_identity_type2:str)->str:
        if (collection_identity_type1 == cn.STRUCTURAL_IDENTITY_TYPE_NOT) \
                or (collection_identity_type2 == cn.STRUCTURAL_IDENTITY_TYPE_NOT):
            return cn.STRUCTURAL_IDENTITY_TYPE_NOT
        if (collection_identity_type1 == cn.STRUCTURAL_IDENTITY_TYPE_WEAK) \
                or (collection_identity_type2 == cn.STRUCTURAL_IDENTITY_TYPE_WEAK):
            return cn.STRUCTURAL_IDENTITY_TYPE_WEAK
        return cn.STRUCTURAL_IDENTITY_TYPE_STRONG
    
    def __add__(self, other:'NetworkCollection')->'NetworkCollection':
        """
        Union of two network collections. Constructs the correct collection_identity_type.

        Args:
            other (NetworkCollection)

        Returns:
            NetworkCollection: _description_

        Raises:
            ValueError: Common names between collections
        """
        # Error checking
        this_names = set([n.network_name for n in self.networks])
        other_names = set([n.network_name for n in other.networks]) 
        common_names = this_names.intersection(other_names)
        if len(common_names) > 0:
            raise ValueError(f"Common names between collections: {common_names}")
        # Construct the new collection
        networks = copy.deepcopy(self.networks)
        networks.extend(other.networks)
        directory = None
        if self.directory == other.directory:
            directory = self.directory
        return NetworkCollection(networks, directory=directory)
    
    def copy(self)->'NetworkCollection':
        return copy.deepcopy(self)
    
    @classmethod
    def makeRandomCollection(cls, num_species:int=3, num_reaction:int=3, num_network:int=10)->'NetworkCollection':
        """
        Make a collection of random networks according to the specified parameters.

        Args:
            array_size (int, optional): Size of the square matrix
            num_network (int, optional): Number of networks

        Returns:
            NetworkCollection
        """
        networks = [Network.makeRandomNetworkByReactionType(num_species=num_species, num_reaction=num_reaction)
                    for _ in range(num_network)]
        return cls(networks)

    @classmethod
    def makeFromAntimonyDirectory(cls, indir_path:str, 
                batch_size:Optional[float]=None,
                processed_network_names:Optional[List[str]]=None,
                report_interval:Optional[int]=None)->'NetworkCollection':
        """Creates a NetworkCollection from a directory of Antimony files.

        Args:
            indir_path (str): Path to the antimony model directory
            batch_size (int): Number of files to process. Default is 5. -1 is all.
            max_file (int): Maximum number of files to process
            processed_network_names (List[str]): Names of models already processed
            report_interval (int): Report interval. Default is None (no report)

        Returns:
            NetworkCollection
        """
        if batch_size is None:
            batch_size = 5
        elif batch_size < 0:
            batch_size = float("inf")
        ffiles = os.listdir(indir_path)
        networks = []
        network_names = []
        num_processed = 0
        if processed_network_names is not None:
            network_names = list(processed_network_names)
        for count, ffile in enumerate(ffiles):
            if (num_processed is not None) and (num_processed >= batch_size):  # type: ignore
                break
            if report_interval is not None and count % report_interval == 0:
                is_report = True
            else:
                is_report = False
            network_name = ffile.split('.')[0]
            if network_name in network_names:
                if is_report:
                    print(".", end='')
                continue
            if ffile.endswith(".xml"):
                try:
                    roadrunner = te.loadSBMLModel(os.path.join(indir_path, ffile))
                    clean_antimony_str = roadrunner.getAntimony()
                except:
                    print(f"Could not process {ffile}. File ignored.")
                    continue
                network = Network.makeFromAntimonyStr(clean_antimony_str, roadrunner=roadrunner, network_name=network_name)
            elif any([ffile.endswith(ext) for ext in ANTIMONY_EXTS]) or len(ffile.split('.')) == 1:
                path = os.path.join(indir_path, ffile)
                network = Network.makeFromAntimonyFile(path, network_name=network_name)
            else:
                continue
            if (network is None) and (report_interval is not None):
                print(f"Could not process {ffile}")
            if (network is None):
                continue
            networks.append(network)
            num_processed += 1
            if is_report:
                print(f"Processed {count} files.")
        return NetworkCollection(networks, directory=indir_path)

#    def serialize(self)->pd.DataFrame:
#        """Constructs a DataFrame from a NetworkCollection
#
#        Returns:
#            pd.DataFrame: See SERIALIZATION_NAMES
#               DataFrame.metadata: Dictionary of metadata
#                    "directory": Directory of the Antimony files
#        """
#        sers = [n.serialize() for n in self.networks]
#        return pd.concat(sers, axis=1).transpose()
#
#    @classmethod 
#    def deserialize(cls, df:pd.DataFrame)->'NetworkCollection':
#        """Deserializes a DataFrame to a NetworkCollection
#
#        Args:
#            df: pd.DataFrame
#
#        Returns:
#            NetworkCollection
#        """
#        networks = []
#        for _, row in df.iterrows():
#            import pdb; pdb.set_trace()
#            network = Network.deserialize(row)
#            networks.append(network)
#        return NetworkCollection(networks)

    def serialize(self)->str:
        """Constructs a json string that serializes the object.

        Returns:
            str
        """
        dct = {cn.S_ID: self.__class__.__name__,
               cn.S_NETWORKS: [n.serialize() for n in self.networks],
               cn.S_DIRECTORY: self.directory}
        return json.dumps(dct)

    @classmethod 
    def deserialize(cls, serialization_str:str)->'NetworkCollection':
        """Deserializes a json string to a NetworkCollection

        Args:
            str

        Returns:
            NetworkCollection
        """
        dct = json.loads(serialization_str)
        if not cls.__name__ in dct[cn.S_ID]:
            raise ValueError(f"Expected {cls} but got {dct[cn.S_ID]}")
        networks = [Network.deserialize(n) for n in dct[cn.S_NETWORKS]]
        if cn.S_DIRECTORY in dct:
            directory = dct[cn.S_DIRECTORY]
        else:
            directory = None
        return NetworkCollection(networks, directory=directory)
    
    @classmethod
    def dataframeToJson(cls, df:pd.DataFrame)->str:
        """Converts a DataFrame to a json string.

        Args:
            df (pd.DataFrame)

        Returns:
            str
        """
        dct = {cn.S_ID: str(cls),
               cn.S_NETWORKS: [Network.seriesToJson(row) for _, row in df.iterrows()],
               cn.S_ANTIMONY_DIRECTORY: None,
        }
        return json.dumps(dct)


import pySubnetSB.constants as cn  # type: ignore
import pySubnetSB.parallel_subnet_finder_worker as psfw  # type: ignore
from pySubnetSB.network import Network  # type: ignore

import os
import json
import pandas as pd
import unittest

IGNORE_TEST = False
IS_PLOT =  False
SIZE = 10
NUM_NETWORK = 10    
BASE_CHECKPOINT_PATH = os.path.join(cn.TEST_DIR, "test_subnet_finder_checkpoint.csv")
WORKER0_CHECKPOINT_PATH = psfw.WorkerCheckpointManager.makeWorkerCheckpointPath(BASE_CHECKPOINT_PATH, 0)
WORKER1_CHECKPOINT_PATH = psfw.WorkerCheckpointManager.makeWorkerCheckpointPath(BASE_CHECKPOINT_PATH, 1)
REMOVE_FILES:list = [BASE_CHECKPOINT_PATH, WORKER0_CHECKPOINT_PATH, WORKER1_CHECKPOINT_PATH]


#############################
class TestWorkerCheckpointManager(unittest.TestCase):

    def setUp(self):
        self.remove()
        self.checkpoint_manager = psfw.WorkerCheckpointManager(BASE_CHECKPOINT_PATH, is_report=IS_PLOT)

    def remove(self):
        for ffile in REMOVE_FILES:
            if os.path.exists(ffile):
                os.remove(ffile)

    def tearDown(self):
        self.remove()

    @staticmethod
    def makeDataframe(num_network:int)->pd.DataFrame:
        # Creates a dataframe used by the WorkerCheckpointManager
        reference_networks = [Network.makeRandomNetworkByReactionType(3, is_prune_species=True) for _ in range(num_network)]
        target_networks = [Network.makeRandomNetworkByReactionType(3, is_prune_species=True) for _ in range(num_network)]
        dct = {cn.FINDER_REFERENCE_NETWORK: [str(n) for n in range(num_network)],
            cn.FINDER_INDUCED_NETWORK: [str(n) for n in range(num_network, 2*num_network)]}
        df = pd.DataFrame(dct)
        df[cn.FINDER_REFERENCE_NAME] = [str(n) for n in reference_networks]
        df[cn.FINDER_TARGET_NAME] = [str(n) for n in target_networks]
        df[cn.FINDER_NAME_DCT] = [json.dumps(dict(a=n)) for n in range(num_network)]
        df[cn.FINDER_IS_TRUNCATED] = [False for _ in range(num_network)]
        return df

    def testRecover(self):
        if IGNORE_TEST:
            return
        num_network = 10
        df = self.makeDataframe(num_network)
        df.loc[0, cn.FINDER_REFERENCE_NETWORK] = ""
        df.loc[0, cn.FINDER_INDUCED_NETWORK] = ""
        self.checkpoint_manager.checkpoint(df)
        result = self.checkpoint_manager.recover()
        self.assertEqual(len(result.full_df), num_network)
        self.assertEqual(len(result.pruned_df), num_network-1)
        self.assertEqual(len(result.processeds), 1)

    # FIXME: test all fields of prune_result
    def testPrune(self):
        if IGNORE_TEST:
            return
        num_network = 10
        df = self.makeDataframe(num_network)
        df.loc[0, cn.FINDER_REFERENCE_NETWORK] = ""
        df.loc[0, cn.FINDER_INDUCED_NETWORK] = ""
        prune_result = self.checkpoint_manager.prune(df)
        self.assertEqual(len(prune_result.reference_names), 1)
        self.assertEqual(len(prune_result.pruned_df), num_network - 1)


if __name__ == '__main__':
    unittest.main(failfast=False)


'''Worker running in a separate process that evaluates the assignments of rows and columns in the target matrix'''

from pySubnetSB.assignment_pair import AssignmentPair # type: ignore

import collections
import numpy as np
from tqdm import tqdm # type: ignore
from typing import Union, Tuple, Optional, List

WorkerResult = collections.namedtuple('WorkerResult', ['is_truncated', 'assignment_pairs'])  


#########################################################
class _Assignment(object):
     
    def __init__(self, array:np.ndarray):
        self.array = array
        self.num_row, self.num_column = array.shape

    def __repr__(self)->str:
        return f"Assignment({self.array})"


#########################################################
class AssignmentEvaluatorWorker(object):
       
    def __init__(self, reference_arr:np.ndarray, target_arr:np.ndarray, max_batch_size:int):
        """
        Args:
            reference_arr (np.ndarray): reference matrix
            target_arr (np.ndarray): target matrix
            row_assignment_arr (np.ndarray): candidate assignments of target rows to reference rows
            column_assignment_arr (np.ndarray): candidate assignments of target columns to reference columns
            comparison_criteria (ComparisonCriteria): comparison criteria
            max_batch_size (int): maximum batch size in units of bytes
        """
        self.reference_arr = reference_arr
        self.num_reference_row, self.num_reference_column = self.reference_arr.shape
        self.target_arr = target_arr   # Reduce memory usage
        self.max_batch_size = max_batch_size

    @classmethod
    def do(cls, reference_arr:np.ndarray, target_arr:np.ndarray, max_batch_size:int,
           row_assignment_arr:np.ndarray, column_assignment_arr:np.ndarray,
           process_idx:int, total_process:int, result_dct:dict,
           max_num_assignment:int=int(1e8),
           is_report:bool=True)->WorkerResult:
        """
        Evaluates the assignments of rows and columns in the target matrix.

        Args:
            reference_arr (np.ndarray): reference matrix
            target_arr (np.ndarray): target matrix
            max_batch_size (int): maximum batch size in units of bytes
            row_assignment_arr (np.ndarray): candidate assignments of target rows to reference rows
            column_assignment_arr (np.ndarray): candidate assignments of target columns to reference columns
            process_idx (int): numerical index of this process
            total_process (int): total number of processes
            result_dct (dict): dictionary of results by process index
            max_num_assignment (int): maximum number of assignments
            is_report (bool): True if reporting progress

        Returns:
            WorkerResult
                List[AssignmentPair]: pairs of row and column assignments that satisfy the comparison criteria
                is_truncated (bool): True if the number of assignments exceeds the maximum
        """
        worker = cls(reference_arr, target_arr, max_batch_size)
        assignment_pairs = worker.evaluateAssignmentArrays(process_idx, total_process, row_assignment_arr,
              column_assignment_arr, is_report=is_report)
        if len(assignment_pairs) > max_num_assignment:
            is_truncated = True
            assignment_pairs = assignment_pairs[:max_num_assignment]
        else:
            is_truncated = False
        worker_result = WorkerResult(is_truncated=is_truncated, assignment_pairs=assignment_pairs)
        result_dct[process_idx] = worker_result
        return worker_result

    def vectorToLinear(self, num_column_assignment:int, row_idx:Union[int, np.ndarray[int]],  # type: ignore
          column_idx:Union[int, np.ndarray[int]])->Union[int, np.ndarray[int]]:  # type: ignore
        """
        Converts a vector address of the candidate pair assignments (row, column) to a linear address.
        The linearlization is column first.

        Args:
            num_column_assignment (int): number of column assignments
            row_idx (int): row index
            column_idx (int): column index

        Returns:
            np.ndarray: linear index
        """
        return row_idx*num_column_assignment + column_idx

    def linearToVector(self, num_column_assignment:int,
          index:Union[int, np.ndarray])->Union[Tuple[int, int], Tuple[np.ndarray, np.ndarray]]:  
        """
        Converts a linear address of the candidate assignments to a vector index.
        Converts a vector representation of the candidate assignments (row, column) to a linear index

        Args:
            num_column_assignment (int): number of column assignments
            index (int): linear index

        Returns:
            row_idx (int): row index
            column_idx (int): column index
        """
        row_idx = index//num_column_assignment
        column_idx = index%num_column_assignment
        return row_idx, column_idx  # type: ignore

    def _makeBatch(self, start_idx:int, end_idx:int, row_assignment:_Assignment, column_assignment:_Assignment,
          big_reference_arr:Optional[np.ndarray]=None)->Tuple[np.ndarray, np.ndarray]:
        """
        Constructs the reference and target matrices for a batch of comparisons. The approach is:
        1. Construct a flattened version of the target matrix for each comparison so that elements are ordered by
            row, then column, then comparison instance.

        Args:
            start_idx (int): start index
            end_idx (int): end index
            row_assignment (Assignment): row assignments
            column_assignment (Assignment): column assignments
            big_reference_arr (np.ndarray): big reference array

        Returns:
            np.ndarray: big_reference_arr
            np.ndarray: big_target_arr
        """
        num_comparison = end_idx - start_idx + 1
        if num_comparison > row_assignment.num_row*column_assignment.num_row:
            raise ValueError("Number of comparisons exceeds the number of assignments")
        num_column_assignment = column_assignment.num_row
        row_assignment_sel_arr, column_assignment_sel_arr = self.linearToVector(
              num_column_assignment, np.array(range(start_idx, end_idx+1)))
        if row_assignment_sel_arr.shape[0] != num_comparison:  # type: ignore
            raise ValueError("Number of comparisons does not match the indices.")
        if column_assignment_sel_arr.shape[0] != num_comparison:  # type: ignore
            raise ValueError("Number of comparisons does not match the indices.")
        # Calculate the index of rows for the flattened target matrix. There is a row value for each
        #   comparison and element of the reference matrix.
        #      Index of the target rows
        row1_idx_arr = row_assignment.array[row_assignment_sel_arr].flatten()
        #      Replicate each index for the number of columns
        row2_idx_arr = np.repeat(row1_idx_arr, self.num_reference_column)
        # Calculate the column for each element of the target flattened matrix
        column1_idx_arr = column_assignment.array[column_assignment_sel_arr]
        column_idx_arr = np.repeat(column1_idx_arr, self.num_reference_row, axis=0).flatten()
        #assert(len(row2_idx_arr) == len(column_idx_arr))
        # Construct the selected parts of the target matrix
        flattened_target_arr = self.target_arr[row2_idx_arr, column_idx_arr]
        big_target_arr = np.reshape(flattened_target_arr,
              (self.num_reference_row*num_comparison, self.num_reference_column))
        #assert(big_target_arr.shape[0] == self.num_reference_row*num_comparison)
        # Construct the reference array
        if big_reference_arr is None:
            big_references = [self.reference_arr.flatten()]*num_comparison
            flattened_big_reference_arr = np.concatenate(big_references, axis=0)
            big_reference_arr = np.reshape(flattened_big_reference_arr,
                (self.num_reference_row*num_comparison, self.num_reference_column))
        return big_reference_arr, big_target_arr

    @staticmethod 
    def compare(big_reference_arr:np.ndarray, big_target_arr:np.ndarray, num_reference_row:int, is_close:bool)->np.ndarray:
        """
        Compares the reference matrix to the target matrix using a vectorized approach.

        Args:
            big_reference_arr: np.ndarray - reference stoichiometry matrix
            big_target_arr: np.ndarray  - target stoichiometry matrix
            num_reference_row: int - number of reference rows
            is_close: bool - True if the comparison uses np.isclose
        
        Returns:
            np.ndarray[bool]: indicates successful (True) or unsuccessful (False) comparisons by assignment pair index
        """
        # Initializations
        num_big_row, num_column = big_reference_arr.shape
        if is_close:
            big_compatible_arr = np.isclose(big_reference_arr, big_target_arr)
        else:
            big_compatible_arr = big_reference_arr == big_target_arr
        # Determine the successful assignment pairs
        big_row_sum = np.sum(big_compatible_arr, axis=1)
        big_row_satisfy = big_row_sum == num_column # Sucessful row comparisons
        num_assignment_pair = num_big_row//num_reference_row
        assignment_pair_satisfy_arr = np.reshape(big_row_satisfy, (num_assignment_pair, num_reference_row))
        assignment_satisfy_arr = np.sum(assignment_pair_satisfy_arr, axis=1) == num_reference_row
        #
        return assignment_satisfy_arr

    def evaluateAssignmentArrays(self, process_num:int, total_process:int, row_assignment_arr:np.ndarray,
          column_assignment_arr:np.ndarray, max_assignment_pair:int=-1, is_report:bool=True)->List[AssignmentPair]:
        """Finds the row and column assignments that satisfy the comparison criteria.
        Uses equality in comparisons (for speed) since it is expected that array elements
        will be 1 byte integers.

        Args:
            process_num (int): numerical index of this process
            total_process (int): total number of processes
            row_assignment_arr (np.ndarray): assignments of target rows to reference rows
            column_assignment_arr (np.ndarray): assignments of target columns to reference columns
            max_assignment_pair (int): maximum number of assignment pairs (no limit if -1)

        Returns:
            List[AssignmentPair]: pairs of row and column assignments that satisfy the comparison criteria
        """
        # Initializations
        num_row_assignment = row_assignment_arr.shape[0]
        num_column_assignment = column_assignment_arr.shape[0]
        num_comparison = num_row_assignment*num_column_assignment
        row_assignment = _Assignment(row_assignment_arr)
        column_assignment = _Assignment(column_assignment_arr)
        # Error checks
        if row_assignment.num_column != self.num_reference_row:
            raise ValueError("Number of reference rows does not match the number of row assignments")
        if column_assignment.num_column != self.num_reference_column:
            raise ValueError("Number of reference columns does not match the number of row assignments")
        # Calculate the number of assignment pair indices in a batch
        bytes_per_comparison = 2*self.reference_arr.itemsize*self.num_reference_row*self.num_reference_column
        max_comparison_per_batch = max(self.max_batch_size//bytes_per_comparison, 1)
        num_batch = num_comparison//max_comparison_per_batch + 1
        # Iterative do the assignments
        assignment_pairs:list = []
        big_reference_arr = None
        #####
        def loop(ibatch:int, big_reference_arr:Optional[np.ndarray]=None):
            # Performs on step in iteration. This structure is used because of tqdm
            start_idx = ibatch*max_comparison_per_batch
            if start_idx >= num_comparison:
                return
            end_idx = min((ibatch+1)*max_comparison_per_batch, num_comparison - 1)
            if end_idx == num_comparison - 1:
                # Last batch must adjust the size of the reference array
                big_reference_arr = None
            big_reference_arr, big_target_arr = self._makeBatch(start_idx, end_idx, row_assignment, column_assignment,
                    big_reference_arr=big_reference_arr)
            assignment_satisfy_arr = self.compare(big_reference_arr, big_target_arr, self.num_reference_row,
                  is_close=False)
            assignment_satisfy_idx = np.where(assignment_satisfy_arr)[0]
            # Add the assignment pairs
            for idx in assignment_satisfy_idx:
                adjusted_idx = idx + start_idx
                row_idx, column_idx = self.linearToVector(num_column_assignment, adjusted_idx)
                assignment_pair = AssignmentPair(row_assignment=row_assignment_arr[row_idx],
                      column_assignment=column_assignment_arr[column_idx])
                assignment_pairs.append(assignment_pair)
            return
        #####
        if (process_num == 0) and is_report:
            for ibatch in tqdm(range(num_batch), unit_scale=max_comparison_per_batch*total_process,
                 desc=" mapping pairs"):
                loop(ibatch, big_reference_arr=big_reference_arr)
                if max_assignment_pair > 0:
                    import pdb; pdb.set_trace()
                    if len(assignment_pairs) >= max_assignment_pair:
                        break
        else:
            for ibatch in range(num_batch):
                loop(ibatch, big_reference_arr=big_reference_arr)
                if max_assignment_pair > 0:
                    if len(assignment_pairs) >= max_assignment_pair:
                        break
        return assignment_pairs


'''Central class in the DISRN algorithm. Does analysis of network structures.'''

from pySubnetSB import constants as cn  # type: ignore
from pySubnetSB import util  # type: ignore
from pySubnetSB.matrix import Matrix  # type: ignore
from pySubnetSB.reaction_constraint import ReactionConstraint  # type: ignore
from pySubnetSB.species_constraint import SpeciesConstraint   # type: ignore
from pySubnetSB.network_base import NetworkBase, AssignmentPair  # type: ignore
from pySubnetSB.assignment_evaluator import AssignmentEvaluator  # type: ignore
from src.pySubnetSB.performance_monitor import PerformanceMonitor  # type: ignore

from memory_profiler import profile  # type: ignore
import pynauty  # type: ignore
import numpy as np
from typing import Optional, List, Tuple, Union


NULL_ARRAY = np.array([])  # Null array
ATTRS = ["reactant_nmat", "product_nmat", "reaction_names", "species_names", "network_name"]
MAX_PREFIX_LEN = 3   # Maximum length of a prefix in the assignment to do a pairwise analysis
IS_ENABLE_PERFORANCE_REPORTING = False


class StructuralAnalysisResult(object):
    # Auxiliary object returned by isStructurallyIdentical

    def __init__(self,
                 assignment_pairs:list[AssignmentPair],
                 is_truncated:Optional[bool]=False,
                 num_species_candidate:int=-1,
                 num_reaction_candidate:int=-1,
                 network:Optional['Network']=None,
                 )->None:
        """
        Args:
            assignment_pairs (list[AssignmentPair]): List of assignment pairs.
            is_trucnated (bool): True if the number of assignments exceeds the maximum number of assignments.
            num_species_candidate (int): Number of species candidates assignments
            num_reaction_candidate (int): Number of reaction candidates assignments.
        """
        self.assignment_pairs = assignment_pairs
        self.is_truncated = is_truncated
        self.num_species_candidate = num_species_candidate
        self.num_reaction_candidate = num_reaction_candidate
        self.network = network

    @property
    def mapping_pairs(self)->List[AssignmentPair]:
        """Same as assignment pairs."""
        return self.assignment_pairs

    @property
    def induced_network(self)->'Network':
        """Induced network from the first assignment pair."""
        return self.makeInducedNetwork()

    def makeInducedNetwork(self, assignment_pair_idx:int=0)->'Network':
        """
        Creates an induced network from the assignment pair.

        Args:
            assignment_pair_idx (int): index of the assignment pair

        Returns:
            str
        """
        if self.network is None:
            raise ValueError("Network is not defined.")
        if len(self.assignment_pairs) <= assignment_pair_idx:
            msg = f'Assignment pair index {assignment_pair_idx} is out of range.'
            msg += f' Max is {len(self.assignment_pairs)}'
            raise ValueError(msg)
        return self.network.makeInducedNetwork(self.assignment_pairs[assignment_pair_idx])

    def __bool__(self)->bool:
        return len(self.assignment_pairs) > 0
    
    def __repr__(self)->str:
        repr = f"StructurallyIdenticalResult(assignment_pairs={self.assignment_pairs};"
        repr += f" is_truncated={self.is_truncated};"
        return repr


class Network(NetworkBase):

    def __init__(self, reactant_arr:Union[np.ndarray, Matrix], 
                 product_arr:Union[np.ndarray, Matrix],
                 reaction_names:Optional[np.ndarray[str]]=None, # type: ignore
                 species_names:Optional[np.ndarray[str]]=None,  # type: ignore
                 network_name:Optional[str]=None)->None:               # type: ignore
        """
        Args:
            reactant_arr (np.ndarray): Reactant matrix.
            product_arr (np.ndarray): Product matrix.
            network_name (str): Name of the network.
            reaction_names (np.ndarray[str]): Names of the reactions.
            species_names (np.ndarray[str]): Names of the species
        """
        if isinstance(reactant_arr, Matrix):
            reactant_arr = reactant_arr.values
        if isinstance(product_arr, Matrix):
            product_arr = product_arr.values
        super().__init__(reactant_arr, product_arr, network_name=network_name,
                            reaction_names=reaction_names, species_names=species_names)
        
    def isEquivalent(self, other)->bool:
        """Same except for the network name.

        Args:
            other (_type_): _description_

        Returns:
            bool: _description_
        """
        if not isinstance(other, self.__class__):
            return False
        return super().isEquivalent(other)

    def __eq__(self, other)->bool:
        """
        Args:
            other (Network): Network to compare to.
        Returns:
            bool: True if equal.
        """
        if not isinstance(other, self.__class__):
            return False
        return super().__eq__(other)
    
    def copy(self):
        """
        Returns:
            Network: Copy of this network.
        """
        return Network(self.reactant_nmat.values.copy(), self.product_nmat.values.copy(),
                        network_name=self.network_name,
                        reaction_names=self.reaction_names,
                        species_names=self.species_names,
                        criteria_vector=self.criteria_vector)
    
    def isIsomorphic(self, target:'Network')->bool:
        """Using pynauty to detect isomorphism of reaction networks.

        Args:
            target (Network)

        Returns:
            bool
        """
        self_graph = self.makePynautyNetwork()
        target_graph = target.makePynautyNetwork()
        return pynauty.isomorphic(self_graph, target_graph)

    def isStructurallyIdentical(self, target:'Network', is_subnet:bool=True, num_process:int=-1,
            max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT,
            max_batch_size:int=cn.MAX_BATCH_SIZE, identity:str=cn.ID_WEAK,
            is_all_valid_assignment:bool=True,
            is_report:bool=True, is_return_if_truncated:bool=True)->StructuralAnalysisResult:
        """
        Determines if the network is structurally identical to another network or subnet of another network.

        Args:
            target (Network): Network to search for structurally identity
            num_process (int, optional): Number of processes (default: -1 is all)
            is_subnets (bool, optional): Consider subsets
            max_num_assignment (int, optional): Maximum number of assignments to search (no limit if negative)
            max_batch_size (int, optional): Maximum batch size
            identity (str, optional): cn.ID_WEAK or cn.ID_STRONG
            is_report (bool, optional): Print report
            is_all_valid_assignment (bool, optional): Return all valid assignments
            is_return_if_truncated (bool, optional): Return if truncation is required

        Returns:
            StructurallyIdenticalResult
        """
        MIN_ASSIGNMENT_FOR_PARALLELISM = 1e7
        monitor = PerformanceMonitor("isStructurallyIdentical: Start", is_enabled=IS_ENABLE_PERFORANCE_REPORTING)
        monitor.add("isStructurallyIdentical/Initialization: Start")
        if self.num_reaction == 0 or target.num_reaction == 0:
              return StructuralAnalysisResult(assignment_pairs=[], 
                  num_reaction_candidate=0,
                  num_species_candidate=0,
                  is_truncated=False)
        # Initialization
        if max_num_assignment < 0:
            log10_max_num_assignment = np.inf
        else:
            log10_max_num_assignment = np.log10(max_num_assignment)
        reference_reactant_nmat, reference_product_nmat = self.makeMatricesForIdentity(identity)
        target_reactant_nmat, target_product_nmat = target.makeMatricesForIdentity(identity)
        monitor.add("isStructurallyIdentical/Initialization: End")
        #####
        def makeAssignmentArr(cls:type)->Tuple[np.ndarray[int], bool, bool]:  # type: ignore
            monitor.add("makeAssignmentArr/makeAssignmentArr: Start")
            reference_constraint = cls(reference_reactant_nmat, reference_product_nmat, is_subnet=is_subnet)
            target_constraint = cls(target_reactant_nmat, target_product_nmat, is_subnet=is_subnet)
            monitor.add("makeAssignmentArr/makeAssignmentArr/Make compatibility collection: Start")
            compatibility_collection = reference_constraint.makeCompatibilityCollection(
                  target_constraint).compatibility_collection
            monitor.add("makeAssignmentArr/makeAssignmentArr/Make compatibility collection: Created")
            compatibility_collection, prune_is_truncated = compatibility_collection.prune(log10_max_num_assignment)
            monitor.add("makeAssignmentArr/makeAssignmentArr/Make compatibility collection: End")
            is_null = compatibility_collection.log10_num_assignment == -np.inf
            if is_null:
                monitor.add("makeAssignmentArr/makeAssignmentArr/Null return")
                return NULL_ARRAY, prune_is_truncated, is_null
            log10_num_assignment = compatibility_collection.log10_num_assignment
            if log10_num_assignment > 0.5*np.log10(max_num_assignment):
                msg = f"makeAssignmentArr/makeAssignmentArr/Assignment array too large[{log10_num_assignment}]"
                monitor.add(msg)
                return NULL_ARRAY, True, is_null
            else:
                msg = f"makeAssignmentArr/makeAssignmentArr/Expand[{log10_num_assignment}]: Start"
                monitor.add(msg)
                # FIXME: The following takes a long time
                assignment_arr, expand_is_truncated = compatibility_collection.expand(max_num_assignment=max_num_assignment)
                if assignment_arr is NULL_ARRAY:
                    return NULL_ARRAY, expand_is_truncated, is_null
                if assignment_arr.ndim < 2:
                    return NULL_ARRAY, expand_is_truncated, is_null
                if assignment_arr.shape[1] == 0:
                    return NULL_ARRAY, expand_is_truncated, is_null
                is_truncated = prune_is_truncated or expand_is_truncated
                monitor.add("makeAssignmentArr/makeAssignmentArr/Expand: End")
                return assignment_arr, is_truncated, is_null
        #####
        # Calculate the compatibility vectors for species and reactions and then construct the assignment arrays
        monitor.add("makeAssignmentArr/Make compatibility vectors: Start")
        species_assignment_arr, is_species_truncated, is_species_null = makeAssignmentArr(SpeciesConstraint)
        reaction_assignment_arr, is_reaction_truncated, is_reaction_null = makeAssignmentArr(ReactionConstraint)
        monitor.add("makeAssignmentArr/Make compatibility vectors: End")
        is_truncated = is_species_truncated or is_reaction_truncated
        # Check if further truncation is required
        size = species_assignment_arr.shape[0], reaction_assignment_arr.shape[0]
        msg = f"makeAssignmentArr/Check further truncation {size}: Start"
        monitor.add(msg)
        num_species_assignment = species_assignment_arr.shape[0]
        num_reaction_assignment = reaction_assignment_arr.shape[0]
        if num_species_assignment*num_reaction_assignment > max_num_assignment:
            is_truncated = True
            if is_return_if_truncated:
                return StructuralAnalysisResult(assignment_pairs=[], 
                  num_reaction_candidate=num_reaction_assignment,
                  num_species_candidate=num_species_assignment,
                  is_truncated=is_truncated)
            else:
                # Truncate the assignment arrays
                species_frac = num_species_assignment/max_num_assignment
                reaction_frac = num_reaction_assignment/max_num_assignment
                species_assignment_arr = util.selectRandom(species_assignment_arr, int(species_frac*max_num_assignment))
                reaction_assignment_arr = util.selectRandom(reaction_assignment_arr, int(reaction_frac*max_num_assignment))
        monitor.add("makeAssignmentArr/Check further truncation: End")
        # Handle null assignment
        monitor.add("makeAssignmentArr/Null assignment: Start")
        is_null = is_species_null or is_reaction_null
        if len(species_assignment_arr) == 0 or len(reaction_assignment_arr) == 0 or is_null:
            return StructuralAnalysisResult(assignment_pairs=[], 
                  num_reaction_candidate=reaction_assignment_arr.shape[0],
                  num_species_candidate=species_assignment_arr.shape[0],
                  is_truncated=is_truncated)
        monitor.add("makeAssignmentArr/Null assignment: End")
        # Evaluate the assignments
        size = species_assignment_arr.shape[0], reaction_assignment_arr.shape[0]
        msg = f"makeAssignmentArr/Evaluate reactant assignments for {size}: Start"
        monitor.add(msg)
        #   Evaluate the assignment pairs for the reactant stoichiometry matrix
        evaluator = AssignmentEvaluator(reference_reactant_nmat.values,
              target_reactant_nmat.values, max_batch_size=max_batch_size)
        total_num_assignment = species_assignment_arr.shape[0]*reaction_assignment_arr.shape[0]
        actual_num_process = num_process if total_num_assignment > MIN_ASSIGNMENT_FOR_PARALLELISM else 1
        result = evaluator.parallelEvaluate(species_assignment_arr, reaction_assignment_arr,
                total_process=actual_num_process, is_report=is_report, max_num_assignment=max_num_assignment)
        is_truncated = is_truncated or result.is_truncated
        reactant_assignment_pairs = result.assignment_pairs
        monitor.add("makeAssignmentArr/Evaluate reactant assignments: End")
        #   Evaluate on product matrices for the assignment pairs found for the reactant matrices
        monitor.add("makeAssignmentArr/Evaluate product assignments: Start")
        evaluator = AssignmentEvaluator(reference_product_nmat.values, target_product_nmat.values,
            max_batch_size=max_batch_size)
        bounded_max_num_assignment = max_num_assignment if is_all_valid_assignment else 1
        assignment_pairs = evaluator.evaluateAssignmentPairs(reactant_assignment_pairs,
              max_num_assignment=bounded_max_num_assignment)
        monitor.add("makeAssignmentArr/Evaluate product assignments: End")
        # Return result
        return StructuralAnalysisResult(assignment_pairs=assignment_pairs,
              num_reaction_candidate=reaction_assignment_arr.shape[0],
              num_species_candidate=species_assignment_arr.shape[0],
              is_truncated=is_truncated,
              network=target)


from pySubnetSB.network_collection import NetworkCollection # type: ignore
from pySubnetSB.processed_network import ProcessedNetwork # type: ignore

import copy
import unittest


IGNORE_TEST = False
IS_PLOT = False
NETWORK_COLLECTION = NetworkCollection.makeRandomCollection(num_network=1)


#############################
# Tests
#############################
class TestprocessedNetwork(unittest.TestCase):

    def setUp(self):
        network = copy.deepcopy(NETWORK_COLLECTION.networks[0])
        self.processed_network = ProcessedNetwork(network.network_name)

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertTrue(isinstance(self.processed_network, ProcessedNetwork))

    def testCopy(self):
        if IGNORE_TEST:
            return
        copy_network = self.processed_network.copy()
        self.assertTrue(copy_network == self.processed_network)

    def testSerializeDeserialize(self):
        if IGNORE_TEST:
            return
        for is_indeterminate in [False, True]:
            self.processed_network.setIndeterminate(is_indeterminate)
            serialization_str = self.processed_network.serialize()
            processed_network = self.processed_network.deserialize(serialization_str)
            self.assertEqual(self.processed_network, processed_network)


if __name__ == '__main__':
    unittest.main()


from pySubnetSB.checkpoint_manager import CheckpointManager # type: ignore
import pySubnetSB.constants as cn  # type: ignore

import numpy as np
import os  # type: ignore
import pandas as pd  # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
PATH = os.path.join(cn.TEST_DIR, "test_checkpoint.csv")
PATH_PAT = os.path.join(cn.TEST_DIR, "test_checkpoint_%d.csv")
IDENTITY_COLUMN = "identity"
REMOVE_FILES:list = [PATH, PATH_PAT % 0, PATH_PAT % 1]
DCT = {IDENTITY_COLUMN: [1, 2, 3], 'a': [1, 2, 3]}
DCT1 = {IDENTITY_COLUMN: [3, 2, 3, 1], 'a': [0, 1, 2, 3]}

#############################
# Tests
#############################
class TestCheckpointManager(unittest.TestCase):

    def setUp(self):
        self.path = PATH
        self.identity_column = IDENTITY_COLUMN
        self.remove()
        self.checkpoint_manager = CheckpointManager(path=self.path, is_report=IS_PLOT)

    def tearDown(self):
        self.remove()

    def remove(self):
        for path in REMOVE_FILES:
            if os.path.exists(path):
                os.remove(path)

    def testConstructor(self):
        if IGNORE_TEST:
            return
        with open(self.path, 'w') as fd:
            fd.write("identity\n")
        self.assertTrue(os.path.isfile(self.path))

    def testCheckpoint(self):
        if IGNORE_TEST:
            return
        def test(size):
            df = pd.DataFrame(range(size))
            self.checkpoint_manager.checkpoint(df)
            self.assertTrue(os.path.isfile(self.path))
            df = pd.read_csv(self.path)
            self.assertEqual(len(df), size)
        #
        test(3)
        test(6)

    def testRecover(self):
        if IGNORE_TEST:
            return
        df1 = pd.DataFrame(DCT)
        self.checkpoint_manager.checkpoint(df1)
        df2 = self.checkpoint_manager.recover()
        df = pd.concat([df1, df2], ignore_index=True)
        self.checkpoint_manager.checkpoint(df)
        df3 = self.checkpoint_manager.recover()
        self.assertTrue(df.equals(df3))

    def testMergeCheckpoint(self):
        if IGNORE_TEST:
            return
        dfs = [pd.DataFrame(DCT), pd.DataFrame(DCT)]
        lengths = [len(df) for df in dfs]
        checkpoint_managers = [CheckpointManager(path=PATH_PAT % n, is_report=IS_PLOT) for n in range(len(dfs))]
        [cm.checkpoint(df) for cm, df in zip(checkpoint_managers, dfs)]
        #
        self.checkpoint_manager.mergeCheckpoint(checkpoint_managers)
        df = self.checkpoint_manager.recover()
        self.assertEqual(len(df), np.sum(lengths))


if __name__ == '__main__':
    unittest.main()


'''Utility functions for the SIRN package.'''
import collections
from functools import wraps, cmp_to_key
import itertools
import json
import numpy as np
import os
import pandas as pd # type: ignore
import psutil  # type: ignore
from typing import List, Tuple, Union, Optional
import warnings
warnings.filterwarnings("ignore")  # Ignore warnings from urllib3
import tempfile
import time
import urllib3  # type: ignore

IS_TIMEIT = False
ArrayContext = collections.namedtuple('ArrayContext', "string, num_row, num_column")
INT_MAX = 1000000

def isInt(val: str)->bool:
    """Determines if a string is an integer.

    Args:
        val (str): A string.

    Returns:
        bool: True if the string is an integer.
    """
    try:
        int(val)
        return True
    except ValueError:
        return False

Statistics = collections.namedtuple("Statistics", "mean std min max count total") 
def calculateSummaryStatistics(arr: Union[list, np.ndarray, pd.Series])->Statistics:
    """Calculates basic statistics for an array.

    Args:
        arr (np.array): An array.

    Returns:
        dict: A dictionary with basic statistics.
    """
    arr = np.array(arr)
    mean = np.mean(arr)
    std = np.std(arr)
    min = float(np.min(arr))
    max = float(np.max(arr))
    count = len(arr)
    total = np.sum(arr)
    return Statistics(mean=mean, std=std, min=min, max=max,
                      count=count, total=total)

def timeit(func):
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        if IS_TIMEIT:
            start_time = time.perf_counter()
        result = func(*args, **kwargs)
        if IS_TIMEIT:
            end_time = time.perf_counter()
            total_time = end_time - start_time
            print(f'Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds')
        return result
    return timeit_wrapper

def repeatArray(array:np.ndarray, num_repeat:int)->np.ndarray:
    """Creates a two dimensional array consisting of num_repeat blocks of the input array.

    Args:
        array (np.array): An array.
        num_repeat (int): Number of times to repeat the array.

    Returns:
        np.array: array.columns X array.rows*num_repeats
    """
    return np.vstack([array]*num_repeat)

def repeatRow(array:np.ndarray, num_repeat:int)->np.ndarray:
    """Creates a two dimensional array consisting of num_repeat repetitions of each
    row of the input array.

    Args:
        array (np.array): An array.
        num_repeats (int): Number of times to repeat the array.

    Returns:
        np.array: array.columns X array.rows*num_repeats
    """
    repeat_arr = np.repeat(array, num_repeat, axis=0)
    return repeat_arr

def arrayProd(arr:np.ndarray)->float:
    """Calculates the product of the elements of an array in log space
    to avoid overflow.

    Args:
        arr (np.array): An array.

    Returns:
        int: The product of the elements of the array.
    """
    return np.exp(np.sum(np.log(arr)))
    
def makeRowOrderIndependentHash(array:np.ndarray)->int:
    """Creates a single integer hash for a 1 or 2 dimensional array
    that depends only on the order of values in the columns (last dimension in the array).
    So, the resulting hash is invariant to permutations of the rows.

    Args:
        array (np.array): An array.

    Returns:
        int: Hash value.
    """
    #####
    def hashRow(row):
        return np.sum(pd.util.hash_array(row))
    #####
    def hash2DArray(array):
        result = []
        for row in array:
            result.append(hashRow(row))
        return np.sum(result)
    #
    if array.ndim == 1:
        return hashRow(array)
    elif array.ndim == 2:
        return hash2DArray(array)
    else:
        raise ValueError("Array must be 1, 2 dimensional.")
    
def deprecatedhashMatrix(matrix:np.ndarray)->int:
    """Creates a single integer hash for a 2 dimensional array.

    Args:
        array (np.array): An 2d array.

    Returns:
        int: Hash value.
    """
    # Encode rows
    is_0_arr = matrix == 0
    is_1_arr = matrix == 1
    is_minus_1_arr = matrix == -1
    is_not_arr = np.logical_or(is_0_arr, is_1_arr)
    is_not_arr = np.logical_or(is_minus_1_arr, is_not_arr)
    is_not_arr = np.logical_not(is_not_arr)
    values = np.sum(is_0_arr, axis=1)
    values += 1000*np.sum(is_1_arr, axis=1)
    values += 1000000*np.sum(is_minus_1_arr, axis=1)
    values += 1000000000*np.sum(is_not_arr, axis=1)
    result = hash(str(pd.util.hash_array(np.sort(values))))
    return result

def hashMatrix(matrix:np.ndarray)->np.int64:
    """Creates a single integer hash for a 2 dimensional array. Matrix cannot have a dimensin > 100.

    Args:
        array (np.array): An 2d array.

    Returns:
        int64: Hash value.
    """
    for dim in matrix.shape:
        if dim > 100:
            raise ValueError("Matrix cannot have a dimension > 100.")
    #####
    def hashRows(array:np.ndarray)->np.int64:
        VALUES = [-2, -1, 0, 1, 2]
        results = []
        for row in array:
            row_encoding = 0
            for idx, val in enumerate(VALUES):
                row_encoding += (10**idx)*np.sum(row == val)
            results.append(row_encoding)
        result_arr = np.sort(np.array(results, dtype=np.int64))
        result = np.int64(np.sum(pd.util.hash_array(result_arr)))
        return result
    #####
    row_hash = hashRows(matrix)
    column_hash = hashRows(matrix.T)
    # Hanlde possible overflow on summation
    try:
        np.seterr(over='raise')
        final_hash = row_hash + column_hash
    except:
        final_hash = np.int64(0.1*row_hash + 0.1*column_hash)
    return final_hash
    
def isArrayLessEqual(left_arr:np.ndarray, right_arr:np.ndarray)->bool:
    """Determines if one array is less than another.

    Args:
        left_arr (np.array): An array.
        right_arr (np.array): An array.

    Returns:
        bool: True if left_arr is less than right_arr.
    """
    if left_arr.shape != right_arr.shape:
        return False
    for left_val, right_val in zip(left_arr, right_arr):
        if left_val < right_val:
            return True
        elif left_val > right_val:
            return False
    return True

def arrayToStr(arr:np.ndarray)->str:
    """Converts an array of integers to a single integer.

    Args:
        arr (np.array): An array of integers.

    Returns:
        int: The integer value of the array.
    """
    return ''.join(map(str, arr))

def arrayToSortedDataFrame(array:np.ndarray)->pd.DataFrame:
    """Converts an array to a sorted pandas DataFrame.

    Args:
        array (np.array): A 2d array.

    Returns:
        pd.DataFrame: A sorted DataFrame.
    """
    sorted_assignment_arr = sorted(array, key=arrayToStr)
    return pd.DataFrame(sorted_assignment_arr)

def pruneArray(array:np.ndarray, max_size:int)->Tuple[np.ndarray, bool]:
    """
    Randomly prunes an array to a maximum size.

    Args:
        array (np.array): A 2d array.
        max_size (int): The maximum number of rows to keep.

    Returns:
        np.array: A pruned array.
    """
    if array.shape[0] <= max_size:
        return array, False
    idxs = np.random.permutation(array.shape[0])[:max_size]
    return array[idxs], True

def array2Context(array:np.ndarray)->ArrayContext:
    array = np.array(array)
    if array.ndim == 1:
        num_column = len(array)
        num_row = 1
    elif array.ndim == 2:
        num_row, num_column = np.shape(array)
    else:
        raise ValueError("Array must be 1 or 2 dimensional.")
    flat_array = np.reshape(array, num_row*num_column)
    str_arr = [str(i) for i in flat_array]
    array_str = "[" + ",".join(str_arr) + "]"
    return ArrayContext(array_str, num_row, num_column)

def string2Array(array_context:ArrayContext)->np.ndarray:
    array = np.array(eval(array_context.string))
    array = np.reshape(array, (array_context.num_row, array_context.num_column))
    return array

def sampleListOfLists(list_of_lists:List[List[int]], num_samples:int)->np.ndarray:
    """Randomly samples the permutations implied by a list of lists.

    Args:
        list_of_lists (List[List[int]]): A list of lists.
        num_samples (int): Number of samples.

    Returns:
        np.ndarray: An array of samples. Columns are instances from list, rows are samples.
    """
    lengths = [len(lst) for lst in list_of_lists]
    arrays = [np.array(lst) for lst in list_of_lists]
    sample_position_arr = np.array([np.random.randint(0, l, num_samples) for l in lengths]).T
    samples = []
    for sample_position in sample_position_arr:
        sample = [a[s] for a, s in zip(arrays, sample_position)]
        samples.append(sample)
    return np.array(samples)

def partitionArray(array:np.ndarray, num_partition:int)->List[np.ndarray]:
    """Partitions an array into num_partitions.

    Args:
        array (np.array): An array.
        num_partitions (int): Number of partitions.

    Returns:
        List[np.ndarray]: A list of partitions.
    """
    actual_num_partition = min(num_partition, len(array))
    partitions:list = [ [] for lst in range(actual_num_partition)]
    [partitions[n%actual_num_partition].append(array[n].tolist()) for n in range(len(array))]
    partitions = [np.array(partition) for partition in partitions]
    return partitions

def serializeDct(dct:dict)->str:
    """Serializes a dictionary.

    Args:
        dct (dict): A dictionary.

    Returns:
        str: A serialized dictionary.
    """
    for key, val in dct.items():
        if isinstance(val, np.ndarray):
            dct[key] = val.tolist()
    return json.dumps(dct)

def selectRandom(array:np.ndarray, num_select:int)->np.ndarray:
    """Randomly selects elements from an array.

    Args:
        array (np.array): An array.
        num_select (int): Number of elements to select.

    Returns:
        np.array: An array of selected elements.
    """
    idxs = np.random.permutation(len(array))[:num_select]
    return array[idxs]

def getDefaultSpeciesNames(num_species:int)->np.ndarray:
    """Creates default names for species.

    Args:
        num_species (int)

    Returns:
        np.ndarray[str]: species names
    """
    return np.array([f"S{i}" for i in range(num_species)])


def getDefaultReactionNames(num_reaction:int)->np.ndarray:
    """Creates default names for reaction.

    Args:
        num_reaction (int)

    Returns:
        np.ndarray[str]: reaction names
    """
    return np.array([f"J{i}" for i in range(num_reaction)])

def encodeIntPair(int1:Union[int, np.ndarray], int2:Union[int, np.ndarray],
      max_int:int=INT_MAX)->Union[np.int64, np.ndarray]:
    """Encodes a pair of integers as a single integer.

    Args:
        int1 (int | p.ndarray): An integer.
        int2 (int | np.ndarray): An integer.
        max_int (int): The maximum integer value. Defaults to 1e6

    Returns:
        int | np.ndarray: An encoded integer.
    """
    if isinstance(int1, np.ndarray):
        is_int = False
        arr1:np.ndarray = int1.astype(np.int64)
        arr2:np.ndarray = int2.astype(np.int64)  # type: ignore
    else:
        is_int = True
        arr1 = np.array([int1]).astype(np.int64)
        arr2 = np.array([int2]).astype(np.int64)
    if any(arr1 > max_int) or any(arr2 > max_int):
        raise ValueError(f"Integers must be less than {max_int}.")
    result = arr1*max_int + arr2
    if is_int:
        return result[0]
    else:
        return result

def decodeIntPair(encoded_int:Union[np.int64, np.ndarray],
      max_int:int=INT_MAX)->Tuple[Union[int, np.ndarray], Union[int, np.ndarray]]:
    """Encodes a pair of integers as a single integer.

    Args:
        encoded_int(int64 | np.ndarray): An integer.
        max_int (int): The maximum integer value. Defaults to 1e6

    Returns:
        int: An encoded integer.
    """
    int2 = encoded_int % max_int
    if isinstance(int2, np.ndarray):
        int2 = [int(i) for i in int2] # type: ignore
    else:
        int2 = int(int2)  # type: ignore
    int1 = (encoded_int - int2)/max_int
    if isinstance(int1, np.ndarray):
        int1 = [int(i) for i in int1] # type: ignore
    else:
        int1 = int(int1)  # type: ignore
    return int1, int2  # type: ignore

def getAllSubsets(a_list)->list:
    """Generates all subsets of a list.

    Args:
        a_list (list): A list.

    Returns:
        list: A list of all subsets.
    """
    result:list = [[]]
    if len(a_list) == 0:
        return result
    #
    for size in range(1, len(a_list)+1):
        for subset in itertools.combinations(a_list, size):
            result.append(list(subset))
    # 
    return result

def getStringsFromURL(url:str)->List[str]:
    """Downloads a file from a URL and convert to list of strings.

    Args:
        url (str): URL of the file.

    Returns:
        List[str]: List of strings.
    """
    http = urllib3.PoolManager()
    resp = http.request("GET", url)
    if resp.status != 200:
        raise ValueError(f"Failed to download file from {url}")
    return resp.data.decode("utf-8").split("\n")

def makeLocalFileFromURL(url:str, local_file:Optional[str]=None)->str:
    """Downloads a file from a URL and saves it locally.

    Args:
        url (str): URL of the file.
        local_file (str): Local file name.

    Returns:
        str: Local file name.
    """
    http = urllib3.PoolManager()
    resp = http.request("GET", url)
    if resp.status != 200:
        raise ValueError(f"Failed to download file from {url}")
    strings = getStringsFromURL(url)
    if local_file is None:
        fp = tempfile.NamedTemporaryFile(delete=False)
        bytes = '\n'.join(strings).encode("utf-8")
        fp.write(bytes)
        local_file = fp.name
        fp.close()
    else:
        with open(local_file, "w") as fp:  # type: ignore
            fp.write(resp.data)
    return local_file

def getMemoryUsage():
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return mem_info.rss # in bytes


import numpy as np
import os


PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))
for _ in range(2):
    PROJECT_DIR = os.path.dirname(PROJECT_DIR)
    if ("pySubnetSB" in PROJECT_DIR) and (not "src" in PROJECT_DIR):
        break
MODEL_DIR = os.path.join(PROJECT_DIR, 'models')
EXAMPLE_DIR = os.path.join(PROJECT_DIR, 'examples')
DATA_DIR = os.path.join(PROJECT_DIR, 'data')
AUXILIARY_DATA_DIR = os.path.join(PROJECT_DIR, 'auxiliary_data')
SCRIPT_DIR = os.path.join(PROJECT_DIR, 'scripts')
TEST_DIR = os.path.join(PROJECT_DIR, 'src', 'pySubnetSB_tests')
PLOT_DIR = os.path.join(PROJECT_DIR, 'plots')
OSCILLATOR_DIRS = [
        "Oscillators_May_28_2024_8898",
        "Oscillators_June_9_2024_14948",
        "Oscillators_June_10_A_11515",
        "Oscillators_June_10_B_10507",
        "Oscillators_June_11_10160",
        "Oscillators_June_11_A_2024_6877",
        "Oscillators_June_11_B_2024_7809",
        "Oscillators_DOE_JUNE_10_17565",
        "Oscillators_DOE_JUNE_12_A_30917",
        "Oscillators_DOE_JUNE_12_B_41373",
        "Oscillators_DOE_JUNE_12_C_27662",
        ]
# File names
SERIALIZATION_FILE = "collection_serialization.txt"
# Oscillator project
OSCILLATOR_PROJECT = os.path.dirname(os.path.abspath(__file__))
for _ in range(3):
    OSCILLATOR_PROJECT = os.path.dirname(OSCILLATOR_PROJECT)
OSCILLATOR_PROJECT = os.path.join(OSCILLATOR_PROJECT, 'OscillatorDatabase')
# Constants
NONE = "none"
MAX_NUM_ASSIGNMENT = 1e9  # Maximum number of permutations to search
MAX_NUM_MAPPING_PAIR = MAX_NUM_ASSIGNMENT
MAX_BATCH_SIZE = int(1e5)  # Matrix memory in bytes used in a comparison batch in AssignmentEvaluator
STRUCTURAL_IDENTITY = "structural_identity"
UNKNOWN_STRUCTURAL_IDENTITY_NAME = "*" # Used for networks whose structural identity cannot be determined
NETWORK_DELIMITER = "---"
NETWORK_NAME_PREFIX_KNOWN = "!"
NETWORK_NAME_PREFIX_UNKNOWN = "?"
NETWORK_NAME_SUFFIX = "_"
NULL_STR = ""
IDENTITY_PREFIX_STRONG = "+"
IDENTITY_PREFIX_WEAK = "-"
NUM_HASH = 'num_hash'
MAX_HASH = 'max_hash'
OTHER_HASH = 1
# Numerical constants
# Network matrices
#   Matrix type
MT_STOICHIOMETRY = 'mt_stoichiometry'
MT_SINGLE_CRITERIA = 'mt_single_criteria'
MT_PAIR_CRITERIA = 'mt_pair_criteria'
MT_LST = [MT_STOICHIOMETRY, MT_SINGLE_CRITERIA, MT_PAIR_CRITERIA]
#   Orientation
OR_REACTION = 'or_reaction'
OR_SPECIES = 'or_species'
OR_LST = [OR_REACTION, OR_SPECIES]
#   Participant
PR_REACTANT = 'pr_reactant'
PR_PRODUCT = 'pr_product'
PR_LST = [PR_REACTANT, PR_PRODUCT]
#   Identity
ID_WEAK = 'id_weak'
ID_STRONG = 'id_strong'
ID_LST = [ID_WEAK, ID_STRONG]
""" NETWORK_NAME = 'network_name'
REACTANT_ARRAY_STR = 'reactant_array_str'
PRODUCT_ARRAY_STR = 'product_array_str'
NUM_SPECIES = 'num_species'
NUM_REACTION = 'num_reaction'
CRITERIA_ARRAY_STR = 'boundary_array_str'
CRITERIA_ARRAY_LEN = 'boundary_array_len'
SERIALIZATION_NAMES = [NETWORK_NAME, REACTANT_ARRAY_STR, PRODUCT_ARRAY_STR, SPECIES_NAMES,
                       REACTION_NAMES, NUM_SPECIES, NUM_REACTION] """
SPECIES_NAMES = 'species_names'
REACTION_NAMES = 'reaction_names'
CRITERIA_BOUNDARY_VALUES = [-2, -1, 0, 1, 2]
# Serialization
S_ANTIMONY_DIRECTORY = "s_antimony_directory"
S_ASSIGNMENT_COLLECTION = 's_assignment_collection'
S_ASSIGNMENT_PAIR = 's_assignment_pair'
S_BOUNDARY_VALUES = "s_boundary_values"
S_PROCESSED_NETWORKS = "s_processed_networks"
S_COLUMN_DESCRIPTION = "s_column_description"
S_COLUMN_NAMES = "s_column_names"
S_CONSTRAINT_PAIR_DCT = 's_constraint_pair_dct'
S_CRITERIA_VECTOR = "s_criteria_vector"
S_DICT = "s_dict"
S_DIRECTORY = "s_directory"
S_HASH_VAL = "s_hash_val"
S_ID = "s_id"  # Class being serialized
S_IDENTITY = "s_identity"
S_IS_INDETERMINATE = 's_is_indeterminate'
S_MODEL_NAME = "s_model_name"
S_NETWORKS = "s_networks"
S_NETWORK_NAME = "s_network_name"
S_NUM_COLUMN = "s_num_column"
S_NUM_REACTION = "s_num_reaction"
S_NUM_ROW = "s_num_row"
S_NUM_SPECIES = "s_num_species"
S_PROCESSING_TIME = 's_processing_time'
S_PRODUCT_LST = "s_product_lst"
S_PRODUCT_NMAT = "s_product_nmat"  # Named matrix
S_REACTANT_LST = "s_reactant_lst"
S_REACTANT_NMAT = "s_reactant_NMAT"
S_REACTION_ASSIGNMENT_LST = "s_reaction_assignment_lst"
S_REACTION_NAMES = "s_reaction_names"
S_ROW_DESCRIPTION = "s_row_description"
S_ROW_NAMES = "s_row_names"
S_REFERENCE = "s_reference"
S_SPECIES_ASSIGNMENT_LST = "s_species_assignment_lst"
S_SPECIES_NAMES = "s_species_names"
S_TARGET = "s_target"
S_VALUES = "s_values"
# Finder dataframe columns
FINDER_WORKER_IDX = "worker_idx"
FINDER_REFERENCE_NAME = "reference_name"
FINDER_REFERENCE_IDX = "reference_idx"
FINDER_TARGET_NAME = "target_name"
FINDER_TARGET_IDX = "target_idx"
FINDER_REFERENCE_NETWORK = "reference_network"
FINDER_INDUCED_NETWORK = "induced_network"
FINDER_NAME_DCT = "name_dct"  # Dictionary of mapping of target names to reference names for species and reactions
FINDER_NUM_ASSIGNMENT_PAIR = "num_assignment_pair"
FINDER_NUM_MAPPING_PAIR = "num_mapping_pair"   # Same as assignment pair
FINDER_IS_TRUNCATED = "is_truncated"
FINDER_DATAFRAME_COLUMNS = [FINDER_REFERENCE_NAME,  FINDER_TARGET_NAME,  FINDER_REFERENCE_NETWORK,
      FINDER_INDUCED_NETWORK,  FINDER_NAME_DCT, FINDER_NUM_ASSIGNMENT_PAIR, FINDER_IS_TRUNCATED,
      FINDER_NUM_MAPPING_PAIR]
# Data columns
D_MODEL_NAME = "model_name"
D_NUM_ITERATION = "num_iteration"
D_REFERENCE_MODEL = "reference_model"
D_TARGET_MODEL = "target_model"
D_NUM_SPECIES = "num_species"
D_NUM_REACTION = "num_reaction"
D_PROBABILITY_OF_OCCURRENCE_WEAK = "probability_of_occurrence_weak"
D_TRUNCATED_WEAK = "truncated_weak"
D_MEAN_TRUNCATED = "mean_truncated"
D_MEAN_PROBABILITY = "mean_probability"
D_STD_PROBABILITY = "std_probability"
D_PROBABILITY_OF_OCCURRENCE_STRONG = "probability_of_occurrence_strong"
D_NUM_REPLICATION = "num_replication"
D_TRUNCATED_STRONG = "truncated_strong"
D_IS_BOUNDARY_NETWORK = "is_boundary_network"
# Data paths
BIOMODELS_SERIALIZATION_PATH = os.path.join(DATA_DIR, "biomodels_serialized.txt")   # Serialized BioModels
OSCILLATORS_SERIALIZATION_PATH = os.path.join(DATA_DIR, "oscillators_serialized.txt")   # Serialized BioModels
FULL_BIOMODELS_STRONG_PATH = os.path.join(DATA_DIR, "full_biomodels_strong.csv")
FULL_BIOMODELS_WEAK_PATH = os.path.join(DATA_DIR, "full_biomodels_weak.csv")
SUBNET_BIOMODELS_STRONG_PRELIMINARY_PATH = os.path.join(AUXILIARY_DATA_DIR, "subnet_biomodels_strong_preliminary.csv")
SUBNET_BIOMODELS_WEAK_PRELIMINARY_PATH = os.path.join(AUXILIARY_DATA_DIR, "subnet_biomodels_weak_preliminary.csv")
SUBNET_BIOMODELS_STRONG_PATH = os.path.join(DATA_DIR, "subnet_biomodels_strong.csv")
SUBNET_BIOMODELS_WEAK_PATH = os.path.join(DATA_DIR, "subnet_biomodels_weak.csv")
BIOMODELS_SUMMARY_PRELIMINARY_PATH = os.path.join(AUXILIARY_DATA_DIR, "biomodels_summary_preliminary.csv")
BIOMODELS_SUMMARY_PATH = os.path.join(DATA_DIR, "biomodels_summary.csv")



'''Creates an Antimony model from a reference.'''

import tellurium as te  # type: ignore

ANT = "ant"
XML = "xml"

def makeAntimony(model_reference:str)->str:
    """
    Creates a roadrunner instance from a model reference.

    Parameters
    ----------
    model_reference: 
        URL (http:...)
        Antimony file (extension is .ant)
        XML file (extension is .xml)
        XML string
        Antimony string
    
    Returns
    -------
    ExtendedRoadrunner object
    """
    if not isinstance(model_reference, str):
        raise ValueError("Invalid model reference")
    #
    if model_reference[0:4] == "http":
        # URL
        roadrunner = te.loadSBMLModel(model_reference)
    else:
        parts = model_reference.split(".")
        if len(parts) == 2:
            # XML file
            if parts[1] == XML:
                roadrunner = te.loadSBMLModel(model_reference)
            elif parts[1] == ANT:
                # Antimony file
                roadrunner = te.loadAntimonyModel(model_reference)
            else:
                # Assume string for antimony model
                if len(model_reference) > 10:
                    if XML in model_reference[0:10]:
                        roadrunner = te.loads(model_reference)
                else:
                    # Possible Antimony string
                    return model_reference
        else:
            if XML in model_reference[0:10]:
                roadrunner = te.loads(model_reference)
            else:
                roadrunner = te.loada(model_reference)
    return roadrunner.getAntimony()



from pySubnetSB.matrix import Matrix # type: ignore

import copy
import numpy as np
import unittest


IGNORE_TEST = False
IS_PLOT = False
MAT = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])


#############################
# Tests
#############################
class TestMatrix(unittest.TestCase):

    def setUp(self):
        self.array = copy.copy(MAT)
        self.matrix = Matrix(self.array)

    def testConstructor(self):
        if IGNORE_TEST:
            return
        matrix = Matrix(self.array)
        self.assertTrue(np.all(matrix.values== self.array))

    def testMakeTrinaryMatrix(self):
        if IGNORE_TEST:
            return
        for _ in range(100):
            try:
                arr = Matrix.makeTrinaryMatrix(3, 3, prob0=0.9)
            except RuntimeError:
                continue
            arr_sq = arr*arr
            is_nozerorow = np.all(arr_sq.sum(axis=1) > 0)
            is_nozerocol = np.all(arr_sq.sum(axis=0) > 0)
            self.assertTrue(is_nozerorow and is_nozerocol)
            break

    def testRandomize(self):
        #if IGNORE_TEST:
        #    return
        matrix = Matrix(Matrix.makeTrinaryMatrix(4, 5, prob0=0.2))
        row_triples = []
        for idx in range(matrix.num_row):
            triple = (np.sum(matrix.values[idx, :] < 0),
                      np.sum(matrix.values[idx, :] == 0),
                      np.sum(matrix.values[idx, :] > 0 ))
            row_triples.append(triple)


        

if __name__ == '__main__':
    unittest.main()